# -*- encoding:utf-8 -*-
"""
    统计相关工具模块
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

try:
    from collections import OrderedDict
except ImportError:
    from ..ExtBu.odict import OrderedDict
from collections import namedtuple, Iterable

import logging
import math
import functools

import numpy as np
import pandas as pd
import scipy.stats as scs
from numpy import linalg as la
from sklearn.metrics.pairwise import euclidean_distances, manhattan_distances, cosine_distances

from ..CoreBu import ABuEnv
from ..UtilBu import ABuScalerUtil
from ..CoreBu.ABuFixes import six
from ..CoreBu.ABuPdHelper import pd_rolling_mean

__author__ = '阿布'
__weixin__ = 'abu_quant'


g_euclidean_safe = True


# noinspection PyUnresolvedReferences
def _distance_matrix(distance_func, df, scale_end=True, to_similar=False):
    """
    非两两distance计算，限制只有一个矩阵的输入，且输入必须为pd.DataFrame or np.array or 多层迭代序列[[],[]]
    :param distance_func: 计算距离的方法
    :param df: pd.DataFrame or np.array or 多层迭代序列[[],[]], 之所以叫df，是因为在内部会统一转换为pd.DataFrame
    :param scale_end: 对结果矩阵进行标准化处理
    :param to_similar: 是否进行后置输出转换similar值
    :return: distance_matrix，pd.DataFrame对象
    """

    if not callable(distance_func):
        raise TypeError('distance_func must callable!!!')

    if isinstance(df, np.ndarray):
        # 把np.ndarray转DataFrame，便统一处理
        df = pd.DataFrame(df)

    if not isinstance(df, pd.DataFrame):
        if all(isinstance(arr_item, Iterable) for arr_item in df):
            # 如果子序列的元素也都是可以迭代的，那么先转np.array，然后再DataFrame
            df = pd.DataFrame(np.array(df))
        else:
            raise TypeError('df must pd.DataFrame object!!!')
    # 做列的distance所以df.T
    distance = distance_func(df.T)

    if scale_end:
        # TODO 这里需要可以设置标准化使用的方法，暂时都使用scaler_mm
        distance = ABuScalerUtil.scaler_mm(distance)
        if to_similar:
            # 只有scale_end的情况下to_similar才会生效，否则没有意义
            distance = 1 - distance

    # 将计算结果的distance转换为pd.DataFrame对象，行和列索引都使用df.columns
    distance_df = pd.DataFrame(distance, index=df.columns, columns=df.columns)
    return distance_df


def _distance_xy(distance_func, x, y):
    """
    两两distance计算，输入x, y
    :param distance_func: 计算距离的方法
    :param x: 可迭代序列
    :param y: 可迭代序列
    :return: float数值
    """
    distance = distance_func(x, y)
    if isinstance(distance, float):
        return distance

    return distance_func(x, y)[0][0]


def euclidean_distance_xy(x, y, to_similar=False):
    """
    欧式距离(L2范数)计算两个序列distance, g_euclidean_safe控制是否使用euclidean_distances计算
    还是使用la.norm，效率差别如下：
        euclidean_distances： 10000 loops, best of 3: 128 µs per loop
        la.norm            ： 10000 loops, best of 3: 89.6 µs per loop
    在需要批量且很耗时的情况下切好模式，否则不要切换
    :param x: 可迭代序列
    :param y: 可迭代序列
    :param to_similar: 是否进行后置输出转换similar值
    :return: float数值

    """
    if g_euclidean_safe:
        euclidean = lambda a, b: euclidean_distances(a, b)
    else:
        euclidean = lambda a, b: la.norm(a - b)
    distance = _distance_xy(euclidean, x, y)
    if to_similar:
        # 实际上l1和l2转换similar的值不直观，只能对比使用
        distance = 1.0 / (1.0 + distance)
    return distance


def euclidean_distance_matrix(df, scale_end=True, to_similar=False):
    """
    欧式距离(L2范数): 与euclidean_distance_xy的区别主要是，非两两distance计算，只有一个矩阵的输入，
    且输入必须为pd.DataFrame or np.array or 多层迭代序列[[],[]], 注意需要理解数据的测距目的来分析
    是否需要进行scale_start，进行和不进行scale_start的结果将完全不一样，在功能需求及数据理解的情况下
    选择是否进行scale_start

            input:

                        tsla	bidu	noah	sfun	goog	vips	aapl
            2014-07-25	223.57	226.50	15.32	12.110	589.02	21.349	97.67
            2014-07-28	224.82	225.80	16.13	12.450	590.60	21.548	99.02
            2014-07-29	225.01	220.00	16.75	12.220	585.61	21.190	98.38
            ...	...	...	...	...	...	...	...
            2016-07-22	222.27	160.88	25.50	4.850	742.74	13.510	98.66
            2016-07-25	230.01	160.25	25.57	4.790	739.77	13.390	97.34
            2016-07-26	225.93	163.09	24.75	4.945	740.92	13.655	97.76

            ABuStatsUtil.euclidean_distance_matrix(cc, scale_start=True)

            output:

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.4086	0.7539	0.7942	0.4810	0.7638	0.3713
            bidu	0.4086	0.0000	0.7732	0.7047	0.6185	0.6161	0.4184
            noah	0.7539	0.7732	0.0000	0.7790	0.7174	0.6957	0.7425
            sfun	0.7942	0.7047	0.7790	0.0000	0.9950	0.5422	0.9558
            goog	0.4810	0.6185	0.7174	0.9950	0.0000	1.0000	0.5379
            vips	0.7638	0.6161	0.6957	0.5422	1.0000	0.0000	0.7348
            aapl	0.3713	0.4184	0.7425	0.9558	0.5379	0.7348	0.0000


            ABuStatsUtil.euclidean_distance_matrix(cc, scale_start=False)

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.0781	0.3314	0.3573	0.6527	0.3386	0.1933
            bidu	0.0781	0.0000	0.2764	0.3018	0.7112	0.2827	0.1392
            noah	0.3314	0.2764	0.0000	0.0284	0.9732	0.0140	0.1408
            sfun	0.3573	0.3018	0.0284	0.0000	1.0000	0.0203	0.1674
            goog	0.6527	0.7112	0.9732	1.0000	0.0000	0.9820	0.8369
            vips	0.3386	0.2827	0.0140	0.0203	0.9820	0.0000	0.1481
            aapl	0.1933	0.1392	0.1408	0.1674	0.8369	0.1481	0.0000

    :param df: pd.DataFrame or np.array or 多层迭代序列[[],[]], 之所以叫df，是因为在内部会统一转换为pd.DataFrame
    :param scale_end: 对结果矩阵进行标准化处理
    :param to_similar: 是否进行后置输出转换similar值
    :return: distance_df，pd.DataFrame对象
    """
    return _distance_matrix(euclidean_distances, df, scale_end, to_similar)


# noinspection PyUnresolvedReferences
def manhattan_distances_xy(x, y, to_similar=False):
    """
    曼哈顿距离(L1范数)计算两个序列distance，注意需要理解数据的测距目的来分析
    是否需要进行scale_start，进行和不进行scale_start的结果将完全不一样，在功能需求及数据理解的情况下
    选择是否进行scale_start
    :param x: 可迭代序列
    :param y: 可迭代序列
    :param to_similar: 是否进行后置输出转换similar值
    :return: float数值
    """
    distance = _distance_xy(manhattan_distances, x, y)
    if to_similar:
        # 实际上l1和l2转换similar的值不直观，只能对比使用
        distance = 1.0 / (1.0 + distance)
    return distance


def manhattan_distance_matrix(df, scale_end=True, to_similar=False):
    """
    曼哈顿距离(L1范数): 与manhattan_distances_xy的区别主要是，非两两distance计算，只有一个矩阵的输入，
    且输入必须为pd.DataFrame or np.array or 多层迭代序列[[],[]]，注意需要理解数据的测距目的来分析
    是否需要进行scale_start，进行和不进行scale_start的结果将完全不一样，在功能需求及数据理解的情况下
    选择是否进行scale_start

        eg:
            input:

                        tsla	bidu	noah	sfun	goog	vips	aapl
            2014-07-25	223.57	226.50	15.32	12.110	589.02	21.349	97.67
            2014-07-28	224.82	225.80	16.13	12.450	590.60	21.548	99.02
            2014-07-29	225.01	220.00	16.75	12.220	585.61	21.190	98.38
            ...	...	...	...	...	...	...	...
            2016-07-22	222.27	160.88	25.50	4.850	742.74	13.510	98.66
            2016-07-25	230.01	160.25	25.57	4.790	739.77	13.390	97.34
            2016-07-26	225.93	163.09	24.75	4.945	740.92	13.655	97.76

            ABuStatsUtil.manhattan_distance_matrix(cc, scale_start=True)

            output:

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.3698	0.6452	0.7917	0.4670	0.7426	0.3198
            bidu	0.3698	0.0000	0.5969	0.7056	0.6495	0.5822	0.4000
            noah	0.6452	0.5969	0.0000	0.7422	0.7441	0.6913	0.6896
            sfun	0.7917	0.7056	0.7422	0.0000	0.9236	0.4489	1.0000
            goog	0.4670	0.6495	0.7441	0.9236	0.0000	0.8925	0.5134
            vips	0.7426	0.5822	0.6913	0.4489	0.8925	0.0000	0.7038
            aapl	0.3198	0.4000	0.6896	1.0000	0.5134	0.7038	0.0000


            ABuStatsUtil.manhattan_distance_matrix(cc, scale_start=False)

            output:

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.0640	0.3318	0.3585	0.6415	0.3395	0.1906
            bidu	0.0640	0.0000	0.2750	0.3018	0.6982	0.2827	0.1338
            noah	0.3318	0.2750	0.0000	0.0267	0.9733	0.0124	0.1412
            sfun	0.3585	0.3018	0.0267	0.0000	1.0000	0.0191	0.1680
            goog	0.6415	0.6982	0.9733	1.0000	0.0000	0.9809	0.8320
            vips	0.3395	0.2827	0.0124	0.0191	0.9809	0.0000	0.1489
            aapl	0.1906	0.1338	0.1412	0.1680	0.8320	0.1489	0.000

    :param df: pd.DataFrame or np.array or 多层迭代序列[[],[]], 之所以叫df，是因为在内部会统一转换为pd.DataFrame
    :param scale_end: 对结果矩阵进行标准化处理
    :param to_similar: 是否进行后置输出转换similar值
    :return: distance_df，pd.DataFrame对象
    """
    return _distance_matrix(manhattan_distances, df, scale_end, to_similar)


def cosine_distances_xy(x, y, to_similar=False):
    """
    余弦距离计算两个序列distance，注意需要理解数据的测距目的来分析
    是否需要进行scale_start，进行和不进行scale_start的结果将完全不一样，在功能需求及数据理解的情况下
    选择是否进行scale_start
    :param x: 可迭代序列
    :param y: 可迭代序列
    :param to_similar: 是否进行后置输出转换similar值
    :return: float数值
    """
    distance = _distance_xy(cosine_distances, x, y)
    if to_similar:
        # 余弦距离转换余弦相似度直接减就行
        distance = 1.0 - distance
    return distance


def cosine_distance_matrix(df, scale_end=True, to_similar=False):
    """
    余弦距离: 与cosine_distances_xy的区别主要是，非两两distance计算，只有一个矩阵的输入，
    且输入必须为pd.DataFrame or np.array or 多层迭代序列[[],[]]，注意需要理解数据的测距目的来分析
    是否需要进行scale_start，进行和不进行scale_start的结果将完全不一样，在功能需求及数据理解的情况下
    选择是否进行scale_start

        eg:
            input:

                        tsla	bidu	noah	sfun	goog	vips	aapl
            2014-07-25	223.57	226.50	15.32	12.110	589.02	21.349	97.67
            2014-07-28	224.82	225.80	16.13	12.450	590.60	21.548	99.02
            2014-07-29	225.01	220.00	16.75	12.220	585.61	21.190	98.38
            ...	...	...	...	...	...	...	...
            2016-07-22	222.27	160.88	25.50	4.850	742.74	13.510	98.66
            2016-07-25	230.01	160.25	25.57	4.790	739.77	13.390	97.34
            2016-07-26	225.93	163.09	24.75	4.945	740.92	13.655	97.76


            ABuStatsUtil.cosine_distance_matrix(cc, scale_start=True)

            output:

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.1743	0.4434	0.2945	0.2394	0.4763	0.1266
            bidu	0.1743	0.0000	0.5808	0.2385	0.3986	0.3034	0.1470
            noah	0.4434	0.5808	0.0000	1.0000	0.3411	0.7626	0.2632
            sfun	0.2945	0.2385	1.0000	0.0000	0.7494	0.4448	0.4590
            goog	0.2394	0.3986	0.3411	0.7494	0.0000	0.9717	0.2806
            vips	0.4763	0.3034	0.7626	0.4448	0.9717	0.0000	0.2669
            aapl	0.1266	0.1470	0.2632	0.4590	0.2806	0.2669	0.0000


            ABuStatsUtil.cosine_distance_matrix(cc, scale_start=False)

            output:

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.1743	0.4434	0.2945	0.2394	0.4763	0.1266
            bidu	0.1743	0.0000	0.5808	0.2385	0.3986	0.3034	0.1470
            noah	0.4434	0.5808	0.0000	1.0000	0.3411	0.7626	0.2632
            sfun	0.2945	0.2385	1.0000	0.0000	0.7494	0.4448	0.4590
            goog	0.2394	0.3986	0.3411	0.7494	0.0000	0.9717	0.2806
            vips	0.4763	0.3034	0.7626	0.4448	0.9717	0.0000	0.2669
            aapl	0.1266	0.1470	0.2632	0.4590	0.2806	0.2669	0.0000

    :param df: pd.DataFrame or np.array or 多层迭代序列[[],[]], 之所以叫df，是因为在内部会统一转换为pd.DataFrame
    :param scale_end: 对结果矩阵进行标准化处理
    :param to_similar: 是否进行后置输出转换similar值
    :return: distance_df，pd.DataFrame对象
    """
    return _distance_matrix(cosine_distances, df, scale_end, to_similar)


# TODO distance与similar之间的关系以及模块功能位置


def arr_to_pandas(func):
    """
        函数装饰器：定参数装饰器，非通用，通用转换使用ABuDTUtil中的装饰器
        将被装饰函数中的arr序列转换为pd.DataFrame或者pd.Series
    """

    @functools.wraps(func)
    def wrapper(arr, *arg, **kwargs):

        # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
        if not isinstance(arr, Iterable) or isinstance(arr, six.string_types):
            # arr必须是可以迭代的对象
            raise TypeError('arr not isinstance of Iterable')

        if not isinstance(arr, pd.DataFrame) or isinstance(arr, pd.Series):
            if isinstance(arr, np.ndarray) and len(arr.shape) > 1 and arr.shape[1] > 1:
                # np.ndarray > 1列的转换为pd.DataFrame
                arr = pd.DataFrame(arr)
            elif isinstance(arr, dict):
                # 针对dict转换pd.DataFrame
                arr = pd.DataFrame(arr)
            elif all(isinstance(arr_item, Iterable) for arr_item in arr):
                # 如果子序列的元素也都是可以迭代的，那么先转np.array，然后再DataFrame
                arr = pd.DataFrame(np.array(arr))
            else:
                # 否则序列对象转换为pd.Series
                arr = pd.Series(arr)
        return func(arr, *arg, **kwargs)

    return wrapper


def arr_to_numpy(func):
    """
        函数装饰器：定参数装饰器，非通用，通用转换使用ABuDTUtil中的装饰器
        将被装饰函数中的arr序列转换为np.array
    """

    @functools.wraps(func)
    def wrapper(arr, *arg, **kwargs):
        # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
        if not isinstance(arr, Iterable) or isinstance(arr, six.string_types):
            # arr必须是可以迭代的对象
            raise TypeError('arr not isinstance of Iterable')

        if not isinstance(arr, np.ndarray):
            if isinstance(arr, pd.DataFrame) or isinstance(arr, pd.Series):
                # 如果是pandas直接拿values
                arr = arr.values
            elif isinstance(arr, dict):
                # 针对dict转换np.array
                arr = np.array(list(arr.values())).T
            else:
                arr = np.array(arr)
        return func(arr, *arg, **kwargs)

    return wrapper


@arr_to_pandas
def demean(arr, rolling_window=0, show=False):
    """
        去均值化后处理demean, 如果输入的是np.array进行转换为pd.DataFrame处理，
        被arr_to_pandas装饰，统一输出，且这样使用arr.mean()不需要axis参数区别np.array轴向
        eg:
            cc.head()
                        tsla	bidu	noah	sfun	goog	vips	aapl
            2014-07-25	223.57	226.50	15.32	12.11	589.02	21.349	97.67
            2014-07-28	224.82	225.80	16.13	12.45	590.60	21.548	99.02
            2014-07-29	225.01	220.00	16.75	12.22	585.61	21.190	98.38
            2014-07-30	228.92	219.13	16.83	11.78	587.42	21.185	98.15
            2014-07-31	223.30	216.05	16.06	11.47	571.60	20.550	95.60

            ABuStatsUtil.demean(cc.head())

                        tsla	bidu	noah	sfun	goog	vips	aapl
            2014-07-25	-1.554	5.004	-0.898	0.104	4.17	0.1846	-0.094
            2014-07-28	-0.304	4.304	-0.088	0.444	5.75	0.3836	1.256
            2014-07-29	-0.114	-1.496	0.532	0.214	0.76	0.0256	0.616
            2014-07-30	3.796	-2.366	0.612	-0.226	2.57	0.0206	0.386
            2014-07-31	-1.824	-5.446	-0.158	-0.536	-13.25	-0.6144	-2.164

            ABuStatsUtil.demean(cc.head().values)

                0	1	2	3	4	5	6
            0	-1.554	5.004	-0.898	0.104	4.17	0.1846	-0.094
            1	-0.304	4.304	-0.088	0.444	5.75	0.3836	1.256
            2	-0.114	-1.496	0.532	0.214	0.76	0.0256	0.616
            3	3.796	-2.366	0.612	-0.226	2.57	0.0206	0.386
            4	-1.824	-5.446	-0.158	-0.536	-13.25	-0.6144	-2.164

            tsla.head()

            2014-07-25    223.57
            2014-07-28    224.82
            2014-07-29    225.01
            2014-07-30    228.92
            2014-07-31    223.30

            ABuStatsUtil.demean(tsla.head())

            2014-07-25   -1.554
            2014-07-28   -0.304
            2014-07-29   -0.114
            2014-07-30    3.796
            2014-07-31   -1.824

            ABuStatsUtil.demean(tsla.head().values)

                0
            0	-1.554
            1	-0.304
            2	-0.114
            3	3.796
            4	-1.824

    :param arr: pd.DataFrame or pd.Series or Iterable
    :param rolling_window: 默认＝0，即不使用移动平均做去均值，rolling_window > 0 生效，
                           注意如果rolling_window值过小将导致去均值后不连续，比如5日，10日的
                           结果只能类似close pct_change的结果，如果需求要钝化，最好是两个月以上
                           的交易日数量，user要根据需求，选择使用的参数，
    :param show: 是否可视化去均值后的结果，默认False
    :return:
    """

    if rolling_window > 0:
        # arr_to_pandas装饰器保证了进来的类型不是pd.DataFrame就是pd.Series
        arr_mean = pd_rolling_mean(arr, window=rolling_window, min_periods=1)
        # arr_mean.fillna(method='bfill', inplace=True)
    else:
        arr_mean = arr.mean()

    demean_v = arr - arr_mean
    if show:
        demean_v.plot()
    return demean_v


@arr_to_numpy
def print_stats_sm(arr):
    """
    对arr的统计相关信息输出，并且分别输出数据的实体前面，中间，最后5列, 被arr_to_numpy装饰，
    统一输出，且这样使用arr.max(), arr.min()等不需要axis参数区别

        eg:
                input:

                            tsla	bidu	noah	sfun	goog	vips	aapl
                2014-07-25	223.57	226.50	15.32	12.110	589.02	21.349	97.67
                2014-07-28	224.82	225.80	16.13	12.450	590.60	21.548	99.02
                2014-07-29	225.01	220.00	16.75	12.220	585.61	21.190	98.38
                ...	...	...	...	...	...	...	...
                2016-07-22	222.27	160.88	25.50	4.850	742.74	13.510	98.66
                2016-07-25	230.01	160.25	25.57	4.790	739.77	13.390	97.34
                2016-07-26	225.93	163.09	24.75	4.945	740.92	13.655	97.76

                output:

                ************************************************************************************************************
                shape = (504, 7)
                ************************************************************************************************************
                max  = 776.6
                ************************************************************************************************************
                min = 4.44
                ************************************************************************************************************
                mean = 172.445713435
                ************************************************************************************************************
                ************************************************************************************************************
                head =
                [[ 223.57   226.5     15.32    12.11   589.02    21.349   97.67 ]
                 [ 224.82   225.8     16.13    12.45   590.6     21.548   99.02 ]
                 [ 225.01   220.      16.75    12.22   585.61    21.19    98.38 ]
                 [ 228.92   219.13    16.83    11.78   587.42    21.185   98.15 ]
                 [ 223.3    216.05    16.06    11.47   571.6     20.55    95.6  ]]
                ************************************************************************************************************
                mid =
                [[ 267.2   207.63   26.25    7.44  644.28   21.23  125.16]
                 [ 265.41  206.25   23.4     7.17  623.56   20.25  124.5 ]
                 [ 253.01  197.68   22.16    6.76  627.26   19.99  122.77]
                 [ 264.82  168.03   22.94    6.76  628.     20.2   123.38]
                 [ 263.82  170.01   23.35    6.79  631.93   20.26  122.99]]
                ************************************************************************************************************
                tail =
                [[ 228.36   160.65    25.11     4.9    741.19    13.63    99.96 ]
                 [ 220.5    161.4     25.51     4.79   738.63    13.69    99.43 ]
                 [ 222.27   160.88    25.5      4.85   742.74    13.51    98.66 ]
                 [ 230.01   160.25    25.57     4.79   739.77    13.39    97.34 ]
                 [ 225.93   163.09    24.75     4.945  740.92    13.655   97.76 ]]
    :param arr:
    """
    log_func = logging.info if ABuEnv.g_is_ipython else print
    separator = format('*', '*^108s')
    log_func(separator)
    log_func('shape = ' + str(arr.shape))
    log_func(separator)
    log_func('max  = ' + str(arr.max()))
    log_func(separator)
    log_func('min = ' + str(arr.min()))
    log_func(separator)
    log_func('mean = ' + str(arr.mean()))
    log_func(separator)

    if arr.shape[0] > 5 * 3:
        # 要求数量能满足打三排数据的
        log_func(separator)
        log_func('head = \n' + str(arr[:5]))
        log_func(separator)
        # 使用floor找到中间的index
        mid_ind = int(math.floor(arr.shape[0] / 2))
        # 因为使用floor所以-2, +3
        log_func('mid = \n' + str(arr[mid_ind - 2:mid_ind + 3]))
        log_func(separator)
        log_func('tail = \n' + str(arr[-5:]))


def normality_stats(arr):
    """
    统计信息偏度，峰度，正态分布检测，p-value
        eg:
                input:

                2014-07-25    223.57
                2014-07-28    224.82
                2014-07-29    225.01
                               ...
                2016-07-22    222.27
                2016-07-25    230.01
                2016-07-26    225.93

                output:

                array skew = -0.282635248604699
                array skew p-value = 0.009884539532576725
                array kurt = 0.009313464006726946
                array kurt p-value = 0.8403947352953821
                array norm = NormaltestResult(statistic=6.6961445106692237, pvalue=0.035152053009441256)
                array norm p-value = 0.035152053009441256

                input:

                            tsla	bidu	noah	sfun	goog	vips	aapl
                2014-07-25	223.57	226.50	15.32	12.110	589.02	21.349	97.67
                2014-07-28	224.82	225.80	16.13	12.450	590.60	21.548	99.02
                2014-07-29	225.01	220.00	16.75	12.220	585.61	21.190	98.38
                ...	...	...	...	...	...	...	...
                2016-07-22	222.27	160.88	25.50	4.850	742.74	13.510	98.66
                2016-07-25	230.01	160.25	25.57	4.790	739.77	13.390	97.34
                2016-07-26	225.93	163.09	24.75	4.945	740.92	13.655	97.76

                output:

                array skew = [-0.2826 -0.2544  0.1456  1.0322  0.2095  0.095   0.1719]
                array skew p-value = [ 0.0099  0.0198  0.1779  0.      0.0539  0.3781  0.1124]
                array kurt = [ 0.0093 -0.8414 -0.4205  0.4802 -1.547  -0.9203 -1.2104]
                array kurt p-value = [ 0.8404  0.      0.0201  0.0461  1.      0.      0.    ]
                array norm = NormaltestResult(statistic=array([   6.6961,   52.85  ,    7.2163,   69.0119,    3.7161,
                69.3468, 347.229 ]), pvalue=array([ 0.0352,  0.    ,  0.0271,  0.    ,  0.156 ,  0.    ,  0.    ]))
                array norm p-value = [ 0.0352  0.      0.0271  0.      0.156   0.      0.    ]

    :param arr: pd.DataFrame or pd.Series or Iterable
    """
    log_func = logging.info if ABuEnv.g_is_ipython else print

    log_func('array skew = {}'.format(scs.skew(arr)))
    log_func('array skew p-value = {}'.format(scs.skewtest(arr)[1]))

    log_func('array kurt = {}'.format(scs.kurtosis(arr)))
    log_func('array kurt p-value = {}'.format(scs.kurtosistest(arr)[1]))

    log_func('array norm = {}'.format(scs.normaltest(arr)))
    log_func('array norm p-value = {}'.format(scs.normaltest(arr)[1]))


def print_stats(arr):
    """
    对arr的统计相关信息输出
        eg:
                input:

                            tsla	bidu	noah	sfun	goog	vips	aapl
                2014-07-25	223.57	226.50	15.32	12.110	589.02	21.349	97.67
                2014-07-28	224.82	225.80	16.13	12.450	590.60	21.548	99.02
                2014-07-29	225.01	220.00	16.75	12.220	585.61	21.190	98.38
                ...	...	...	...	...	...	...	...
                2016-07-22	222.27	160.88	25.50	4.850	742.74	13.510	98.66
                2016-07-25	230.01	160.25	25.57	4.790	739.77	13.390	97.34
                2016-07-26	225.93	163.09	24.75	4.945	740.92	13.655	97.76

                output:

                array size = 504
                array min  = [ 143.67  132.37   12.95    4.44  492.55   10.35   90.34]
                array max  = [ 286.04  250.34   37.32   12.52  776.6    30.    133.  ]
                array mean = [ 228.4885  193.4488   23.7362    7.2458  624.3016   19.0181  110.881 ]
                array var  = [  653.509    752.7421    30.1604     3.3109  7310.2084    27.0994
                   135.156 ]
                array std  = [ 25.5638  27.4361   5.4919   1.8196  85.4998   5.2057  11.6257]
                array skew = [-0.2826 -0.2544  0.1456  1.0322  0.2095  0.095   0.1719]
                array kurt = [ 0.0093 -0.8414 -0.4205  0.4802 -1.547  -0.9203 -1.2104]
    :param arr: pd.DataFrame or pd.Series or Iterable
    """
    stats = scs.describe(arr)

    log_func = logging.info if ABuEnv.g_is_ipython else print
    log_func('array size = {}'.format(stats[0]))
    log_func('array min  = {}'.format(stats[1][0]))
    log_func('array max  = {}'.format(stats[1][1]))
    log_func('array mean = {}'.format(stats[2]))
    log_func('array var  = {}'.format(stats[3]))
    log_func('array std  = {}'.format(np.sqrt(stats[3])))
    log_func('array skew = {}'.format(stats[4]))
    log_func('array kurt = {}'.format(stats[5]))


@arr_to_numpy
def stats_dict(arr):
    """
    通过序列构造arr的基础统计信息dict，被arr_to_numpy装饰，统一输出，且这样使用arr.max(), arr.min()等不需要axis参数区别
    :param arr: pd.DataFrame or pd.Series or Iterable
    :return: stats_dict，dict对象
                eg:
                    {'count': 504,
                     'kurtosis': 0.009313464006726946,
                     'max': 286.04000000000002,
                     'mean': 228.48845238095237,
                     'min': 143.66999999999999,
                     'skewness': -0.282635248604699,
                     'std': 25.538448192811927}
    """

    count = arr.shape[0]
    if len(arr.shape) > 1 and arr.shape[1] > 1:
        count = arr.shape[0] * arr.shape[1]

    return {'count': count, 'max': arr.max(), 'min': arr.min(), 'mean': arr.mean(),
            'std': arr.std(), 'skewness': scs.skew(arr), 'kurtosis': scs.kurtosis(arr)
            }


# noinspection PyClassHasNoInit
class AbuMomentsTuple(namedtuple('AbuMomentsTuple',
                                 ('count',
                                  'max',
                                  'min',
                                  'mean',
                                  'std',
                                  'skewness',
                                  'kurtosis'))):
    __slots__ = ()

    def __repr__(self):
        return "count:{}\nmax:{}\nmin:{}\nmean:{}\nstd:{}\nskewness:{}\nkurtosis:{}".format(self.count, self.max,
                                                                                            self.min, self.mean,
                                                                                            self.std, self.skewness,
                                                                                            self.kurtosis)


@arr_to_numpy
def stats_namedtuple(arr):
    """
    通过序列构造arr的基础统计信息dict, 被arr_to_numpy装饰，统一输出，且这样使用arr.max(), arr.min()等不需要axis参数区别
    与stats_dict的区别只是返回namedtuple对象
    :param arr: pd.DataFrame or pd.Series or Iterable
    :return: AbuMomentsTuple对象
                eg:
                    count:504
                    max:286.04
                    min:143.67
                    mean:228.48845238095237
                    std:25.538448192811927
                    skewness:-0.282635248604699
                    kurtosis:0.009313464006726946

    """
    count = arr.shape[0]
    if len(arr.shape) > 1 and arr.shape[1] > 1:
        count = arr.shape[0] * arr.shape[1]

    return AbuMomentsTuple(count, arr.max(), arr.min(), arr.mean(), arr.std(), scs.skew(arr), scs.kurtosis(arr))


def sigmoid(arr):
    """sigmoid序列处理，大的更大，小的更小，分辨不清晰的极值"""
    return 1.0 / (1 + np.exp(-arr))



# -*- encoding:utf-8 -*-
"""量化振幅分析工具图形可视化"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ipywidgets as widgets
from IPython.display import display

from ..UtilBu import ABuKLUtil
from ..WidgetBu.ABuWGToolBase import WidgetToolBase, multi_fetch_symbol_analyse

__author__ = '阿布'
__weixin__ = 'abu_quant'


# noinspection PyUnusedLocal
class WidgetDATool(WidgetToolBase):
    """振幅分析界面"""

    def __init__(self, tool_set):
        """初始化数据分析界面"""
        super(WidgetDATool, self).__init__(tool_set)

        da_list = []
        tip_label1 = widgets.Label(u'分析目标需要在\'分析设置\'中选择', layout=self.label_layout)
        tip_label2 = widgets.Label(u'需要设置多个分析目标进行对比', layout=self.label_layout)
        da_list.append(tip_label1)
        da_list.append(tip_label2)

        date_week_wave_bt = widgets.Button(description=u'交易日震幅对比分析', layout=widgets.Layout(width='98%'),
                                           button_style='info')
        date_week_wave_bt.on_click(self.date_week_wave)
        da_list.append(date_week_wave_bt)

        p_change_stats_bt = widgets.Button(description=u'交易日涨跌对比分析', layout=widgets.Layout(width='98%'),
                                           button_style='info')
        p_change_stats_bt.on_click(self.p_change_stats)
        da_list.append(p_change_stats_bt)

        wave_change_rate_bt = widgets.Button(description=u'振幅统计套利条件分析', layout=widgets.Layout(width='98%'),
                                             button_style='info')
        wave_change_rate_bt.on_click(self.wave_change_rate)
        da_list.append(wave_change_rate_bt)

        date_week_win_bt = widgets.Button(description=u'交易日涨跌概率分析', layout=widgets.Layout(width='98%'),
                                          button_style='info')
        date_week_win_bt.on_click(self.date_week_win)
        da_list.append(date_week_win_bt)

        bcut_change_vc_bt = widgets.Button(description=u'交易日涨跌区间分析(预定区间)', layout=widgets.Layout(width='98%'),
                                           button_style='info')
        bcut_change_vc_bt.on_click(self.bcut_change_vc)
        da_list.append(bcut_change_vc_bt)

        qcut_change_vc_bt = widgets.Button(description=u'交易日涨跌区间分析(不定区间)', layout=widgets.Layout(width='98%'),
                                           button_style='info')
        qcut_change_vc_bt.on_click(self.qcut_change_vc)
        da_list.append(qcut_change_vc_bt)

        self.widget = widgets.VBox(da_list, layout=widgets.Layout(width='58%'))

    @multi_fetch_symbol_analyse
    def qcut_change_vc(self, kl_dict, bt):
        """交易日涨跌区间分析(不定区间)action"""
        display(ABuKLUtil.qcut_change_vc(kl_dict))

    @multi_fetch_symbol_analyse
    def bcut_change_vc(self, kl_dict, bt):
        """交易日涨跌区间分析(预定区间)action"""
        display(ABuKLUtil.bcut_change_vc(kl_dict))

    @multi_fetch_symbol_analyse
    def date_week_win(self, kl_dict, bt):
        """交易日涨跌概率分析action"""
        display(ABuKLUtil.date_week_win(kl_dict))

    @multi_fetch_symbol_analyse
    def wave_change_rate(self, kl_dict, bt):
        """振幅统计套利条件分析action"""
        ABuKLUtil.wave_change_rate(kl_dict)

    @multi_fetch_symbol_analyse
    def date_week_wave(self, kl_dict, bt):
        """交易日震幅对比分析action"""
        display(ABuKLUtil.date_week_wave(kl_dict))

    @multi_fetch_symbol_analyse
    def p_change_stats(self, kl_dict, bt):
        """交易日涨跌对比分析action"""
        ABuKLUtil.p_change_stats(kl_dict)



# coding=utf-8
"""
    symbol模块
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from fnmatch import fnmatch

import numpy as np

from ..CoreBu.ABuEnv import EMarketTargetType, EMarketSubType
from ..CoreBu.ABuFixes import six
from ..UtilBu.ABuStrUtil import to_unicode
from ..UtilBu.ABuLazyUtil import LazyFunc


# noinspection PyProtectedMember
def code_to_symbol(code, rs=True):
    """
    解析code成Symbol,如果code中带有市场编码直接用该市场，否则进行查询所属市场信息，
    如果最后也没有发现symbol所在的市场，会向外raise ValueError
    :param code: str对象，代码 如：300104，sz300104，usTSLA
    :param rs: 没有匹配上是否对外抛异常，默认True
    :return: Symbol对象
    """
    from ..MarketBu.ABuSymbolFutures import AbuFuturesCn, AbuFuturesGB
    from ..MarketBu.ABuSymbolStock import AbuSymbolCN, AbuSymbolUS
    from ..MarketBu.ABuMarket import all_symbol

    if not isinstance(code, six.string_types):
        # code必须是string_types
        raise TypeError('code must be string_types!!!，{} : type is {}'.format(code, type(code)))

    sub_market = None
    market = None
    # 尝试获取市场信息
    head = code[:2].lower()
    if head in [EMarketSubType.SH.value, EMarketSubType.SZ.value] and code[2:].isdigit():
        # 市场信息匹配沪深股，查询子市场
        sub_market = EMarketSubType(head)
        market = EMarketTargetType.E_MARKET_TARGET_CN
    elif head == EMarketTargetType.E_MARKET_TARGET_CN.value and code[2:].isdigit():
        # 没指名A股子市场，使用如hs000001这种形式，需求获取symbol后查询子市场使用AbuSymbolCN().query_symbol_sub_market
        sub_market = EMarketSubType(AbuSymbolCN().query_symbol_sub_market(code[2:]))
        market = EMarketTargetType.E_MARKET_TARGET_CN
    elif head == EMarketSubType.HK.value and (code[2:].isdigit() or code[2:] in Symbol.HK_INDEX):
        # 市场信息匹配港股，查询子市场
        sub_market = EMarketSubType.HK
        market = EMarketTargetType.E_MARKET_TARGET_HK
    elif head == EMarketTargetType.E_MARKET_TARGET_US.value:
        # 市场信息匹配美股，查询子市场，AbuSymbolUS().query_symbol_sub_market
        sub_market = EMarketSubType(AbuSymbolUS().query_symbol_sub_market(code[2:]))
        market = EMarketTargetType.E_MARKET_TARGET_US

    if market is not None and sub_market is not None:
        # 通过stock_code, market, sub_market构建Symbol对象
        stock_code = code[2:].upper()
        return Symbol(market, sub_market, stock_code)

    if code.isdigit():
        if len(code) == 6:
            # 6位全数字，匹配查询a股子市场
            market = EMarketTargetType.E_MARKET_TARGET_CN
            sub_market = EMarketSubType(AbuSymbolCN().query_symbol_sub_market(code))
            return Symbol(market, sub_market, code)
        elif len(code) == 5:
            # 5位全数字，匹配查询港股子市场
            market = EMarketTargetType.E_MARKET_TARGET_HK
            sub_market = EMarketSubType.HK
            return Symbol(market, sub_market, code)
        if rs:
            raise TypeError('cn symbol len = 6, hk symbol len = 5')
    elif code.isalpha() and code in Symbol.HK_INDEX:
        # 全字母且匹配港股大盘'HSI', 'HSCEI', 'HSCCI'
        market = EMarketTargetType.E_MARKET_TARGET_HK
        sub_market = EMarketSubType.HK
        return Symbol(market, sub_market, code)
    elif code.isalpha() and code in all_symbol(EMarketTargetType.E_MARKET_TARGET_TC):
        # 全字母且匹配币类市场''btc', 'ltc'
        market = EMarketTargetType.E_MARKET_TARGET_TC
        sub_market = EMarketSubType.COIN
        return Symbol(market, sub_market, code)
    elif code.isalpha() and code.upper() in all_symbol(EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL):
        # 全字母且匹配国际期货市场
        futures_gb_code = code.upper()
        q_df = AbuFuturesGB().query_symbol(futures_gb_code)
        sub_market = EMarketSubType(q_df.exchange.values[0])
        market = EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL
        return Symbol(market, sub_market, futures_gb_code)
    elif code.isalpha() or (code.startswith('.') and code[1:].isalpha()):
        # 全字母进行美股大盘匹配或者美股股票匹配
        stock_code = code.upper()

        if stock_code in Symbol.K_US_INDEX_FIXES.keys():
            stock_code = Symbol.K_US_INDEX_FIXES[stock_code]

        sub_market = EMarketSubType(AbuSymbolUS().query_symbol_sub_market(code))
        market = EMarketTargetType.E_MARKET_TARGET_US
        return Symbol(market, sub_market, stock_code)
    elif (code[0].isalpha() and code[1:].isdigit()) or (code[:2].isalpha() and code[2:].isdigit()):
        # 匹配国内期货市场symbol
        futures_code = code.upper()
        q_df = AbuFuturesCn().query_symbol(futures_code)
        if q_df is not None:
            sub_market = EMarketSubType(q_df.exchange.values[0])
            market = EMarketTargetType.E_MARKET_TARGET_FUTURES_CN
            return Symbol(market, sub_market, futures_code)

    if rs:
        raise ValueError('arg code :{} format dt support'.format(code))


def __search(market_df, search_match, search_code, search_result, match_key='co_name'):
    """具体搜索执行接口"""

    def __search_whole_code(_match_code):
        _sc_df = market_df[market_df.symbol == _match_code]
        if not _sc_df.empty:
            search_result[_sc_df.symbol.values[0]] = _sc_df[match_key].values[0]
            return True
        return False

    def __search_pinyin_code(_match_code):
        from ..MarketBu.ABuDataFeed import query_symbol_from_pinyin
        # 使用query_symbol_from_pinyin对模糊拼音进行查询
        pinyin_symbol = query_symbol_from_pinyin(_match_code)
        if pinyin_symbol is not None:
            # 需要把拼音code标准化为可查询的code
            search_symbol = code_to_symbol(pinyin_symbol, rs=False)
            if search_symbol is not None:
                _search_code = search_symbol.symbol_code
                sc_df = market_df[market_df.symbol == _search_code]
                if not sc_df.empty:
                    search_result[sc_df.symbol.values[0]] = sc_df[match_key].values[0]

    def __search_fnmatch_info(_search_match):
        # 模糊匹配公司名称信息或者交易产品信息
        mc_df = market_df[market_df[match_key].apply(lambda name:
                                                     fnmatch(to_unicode(name),
                                                             _search_match))]
        if not mc_df.empty:
            for ind in np.arange(0, len(mc_df)):
                mcs = mc_df.iloc[ind]
                search_result[mcs.symbol] = mcs[match_key]

    # 首先全匹配search_code
    if not __search_whole_code(search_code):
        # 如果search_code没有能全匹配成功，使用拼音进行匹配一次
        __search_pinyin_code(search_code)
    # 模糊匹配公司名称或者产品等信息symbol
    __search_fnmatch_info(search_match)


def _us_search(search_match, search_code, search_result):
    """美股市场symbol关键字搜索"""
    from ..MarketBu.ABuSymbolStock import AbuSymbolUS
    __search(AbuSymbolUS().df, search_match, search_code, search_result)


def _cn_search(search_match, search_code, search_result):
    """a股市场symbol关键字搜索"""
    from ..MarketBu.ABuSymbolStock import AbuSymbolCN
    __search(AbuSymbolCN().df, search_match, search_code, search_result)


def _hk_search(search_match, search_code, search_result):
    """港股市场symbol关键字搜索"""
    from ..MarketBu.ABuSymbolStock import AbuSymbolHK
    __search(AbuSymbolHK().df, search_match, search_code, search_result)


def _fcn_search(search_match, search_code, search_result):
    """国内期货symbol关键字搜索"""
    from ..MarketBu.ABuSymbolFutures import AbuFuturesCn
    __search(AbuFuturesCn().futures_cn_df,
             search_match, search_code, search_result, match_key='product')


def _fgb_search(search_match, search_code, search_result):
    """国际期货symbol关键字搜索"""
    from ..MarketBu.ABuSymbolFutures import AbuFuturesGB
    __search(AbuFuturesGB().futures_gb_df,
             search_match, search_code, search_result, match_key='product')


# TODO 币类匹配统一标准规范
def _tc_search(search_match, search_code, search_result):
    if fnmatch(u'比特币', search_match) or 'btc' == search_code:
        search_result['btc'] = u'比特币'
    if fnmatch(u'莱特币', search_match) or 'ltc' == search_code:
        search_result['ltc'] = u'莱特币'


def search_to_symbol_dict(search, fast_mode=False):
    """
    symbol搜索对外接口，全匹配symbol code，拼音匹配symbol，别名匹配，模糊匹配公司名称，产品名称等信息
    eg：
        in：
        search_to_symbol_dict('黄金')
        out：
        {'002155': '湖南黄金',
         '600489': '中金黄金',
         '600547': '山东黄金',
         '600766': '园城黄金',
         '600988': '赤峰黄金',
         'ABX': '巴里克黄金',
         'AU0': '黄金',
         'DGL': '黄金基金-PowerShares',
         'DGLD': '黄金3X做空-VelocityShares',
         'DGP': '黄金2X做多-DB',
         'DGZ': '黄金做空-PowerShares',
         'DZZ': '黄金2X做空-DB',
         'EGO': '埃尔拉多黄金公司',
         'GC': '纽约黄金',
         'GEUR': 'Gartman欧元黄金ETF-AdvisorShares ',
         'GLD': '黄金ETF-SPDR',
         'GLL': '黄金2X做空-ProShares',
         'GYEN': 'Gartman日元黄金ETF-AdvisorShares',
         'HMY': '哈莫尼黄金',
         'IAU': '黄金ETF-iShares',
         'KGC': '金罗斯黄金',
         'LIHR': '利希尔黄金',
         'PRME': '全球黄金地段房地产ETF-First Trust Heitman',
         'RGLD': '皇家黄金',
         'UGL': '黄金2x做多-ProShares',
         'UGLD': '黄金3X做多-VelocityShares'}
    :param search: eg：'黄金'， '58'
    :param fast_mode: 是否尽快匹配，速度优先模式
    :return: symbol dict
    """
    search_symbol_dict = {}
    search = search.lower()
    while len(search_symbol_dict) == 0 and len(search) > 0:
        # 构建模糊匹配进行匹配带通配符的字符串
        search_match = u'*{}*'.format(search)
        # 构建精确匹配或拼音模糊匹配的symbol
        search_symbol = code_to_symbol(search, rs=False)
        search_code = ''
        if search_symbol is not None:
            search_code = search_symbol.symbol_code
        # 对search的内容进行递减匹配
        search = search[:-1]
        # 依次对各个市场进行搜索匹配操作
        _tc_search(search_match, search_code, search_symbol_dict)
        _cn_search(search_match, search_code, search_symbol_dict)
        _us_search(search_match, search_code, search_symbol_dict)
        _hk_search(search_match, search_code, search_symbol_dict)
        _fcn_search(search_match, search_code, search_symbol_dict)
        _fgb_search(search_match, search_code, search_symbol_dict)
        if fast_mode:
            break
    return search_symbol_dict


class Symbol(object):
    """统一所有市场的symbol，统一对外接口对象"""

    # 定义使用的美股大盘
    US_INDEX = ['.DJI', '.IXIC', '.INX']
    # 定义使用的sh大盘
    SH_INDEX = ['000001', '000300']
    # 定义使用的sz大盘
    SZ_INDEX = ['399001', '399006']
    # 定义使用的港股大盘
    HK_INDEX = ['HSI', 'HSCEI', 'HSCCI']

    # 修复美股其它大盘写法匹配，主要为外部user使用code_to_symbol进行标准化使用
    K_US_INDEX_FIXES = {'DJI': '.DJI', 'IXIC': '.IXIC', 'INX': '.INX'}

    def __init__(self, market, sub_market, symbol_code):
        """
        :param market: EMarketTargetType enum对象
        :param sub_market: EMarketSubType enum对象
        :param symbol_code: str对象，不包含市场信息的code
        """
        if isinstance(market, EMarketTargetType) and isinstance(sub_market, EMarketSubType):
            self.market = market
            self.sub_market = sub_market
            self.symbol_code = symbol_code
            self.source = None
        else:
            raise TypeError('market type error')

    def __str__(self):
        """打印对象显示：market， sub_market， symbol_code"""
        return '{}_{}:{}'.format(self.market.value, self.sub_market.value, self.symbol_code)

    __repr__ = __str__

    def __len__(self):
        """对象长度：拼接市场＋子市场＋code的字符串长度"""
        m_symbol = '{}_{}:{}'.format(self.market.value, self.sub_market.value, self.symbol_code)
        return len(m_symbol)

    @LazyFunc
    def value(self):
        """不同市场返回ABuSymbolPd.make_kl_df使用的symbol LazyFunc"""
        if self.market == EMarketTargetType.E_MARKET_TARGET_HK or self.market == EMarketTargetType.E_MARKET_TARGET_US:
            # hk,us eg: usTSLA, hk00836
            return '{}{}'.format(self.market.value, self.symbol_code)
        elif self.market == EMarketTargetType.E_MARKET_TARGET_CN:
            # cn eg: sh000001
            return '{}{}'.format(self.sub_market.value, self.symbol_code)
        # 其它市场直接返回symbol_code
        return self.symbol_code

    def is_a_stock(self):
        """判定是否a股symbol"""
        return self.market == EMarketTargetType.E_MARKET_TARGET_CN

    def is_sh_stock(self):
        """判定是否a股sh symbol"""
        return self.sub_market == EMarketSubType.SH

    def is_sz_stock(self):
        """判定是否a股sz symbol"""
        return self.sub_market == EMarketSubType.SZ

    def is_us_stock(self):
        """判定是否美股 symbol"""
        return self.market == EMarketTargetType.E_MARKET_TARGET_US

    def is_us_n_stock(self):
        """判定是否美股纽约交易所 symbol"""
        return self.sub_market == EMarketSubType.US_N

    def is_us_oq_stock(self):
        """判定是否美股纳斯达克交易所 symbol"""
        return self.sub_market == EMarketSubType.US_OQ

    def is_hk_stock(self):
        """判定是否港股 symbol"""
        return self.market == EMarketTargetType.E_MARKET_TARGET_HK

    def is_sh_index(self):
        """判定是否a股sh 大盘"""
        return self.is_sh_stock() and self.symbol_code in Symbol.SH_INDEX

    def is_sz_index(self):
        """判定是否a股sz 大盘"""
        return self.is_sz_stock() and self.symbol_code in Symbol.SZ_INDEX

    def is_a_index(self):
        """判定是否a股 大盘"""
        return self.is_sh_index() or self.is_sz_index()

    def is_us_index(self):
        """判定是否美股 大盘"""
        return self.is_us_stock() and self.symbol_code in Symbol.US_INDEX

    def is_hk_index(self):
        """判定是否港股 大盘"""
        return self.is_hk_stock() and self.symbol_code in Symbol.HK_INDEX

    def is_index(self):
        """判定是否大盘"""
        return self.is_us_index() or self.is_hk_index() or self.is_a_index()

    def is_futures(self):
        """判定是否期货symbol"""
        return self.market == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN \
            or self.market == EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL

    def is_tc(self):
        """判定是否电子货币symbol"""
        return self.market == EMarketTargetType.E_MARKET_TARGET_TC


class IndexSymbol(object):
    """定义IndexSymbol类，设定大盘指数Symbol对象的规范"""

    # 美股大盘DJI Symbol对象
    DJI = Symbol(EMarketTargetType.E_MARKET_TARGET_US, EMarketSubType.US_N, '.DJI')
    # 美股大盘IXIC Symbol对象
    IXIC = Symbol(EMarketTargetType.E_MARKET_TARGET_US, EMarketSubType.US_N, '.IXIC')
    # 美股大盘INX Symbol对象
    INX = Symbol(EMarketTargetType.E_MARKET_TARGET_US, EMarketSubType.US_N, '.INX')

    # a股sh大盘Symbol对象
    SH = Symbol(EMarketTargetType.E_MARKET_TARGET_CN, EMarketSubType.SH, '000001')
    # a股sz大盘Symbol对象
    SZ = Symbol(EMarketTargetType.E_MARKET_TARGET_CN, EMarketSubType.SZ, '399001')
    # a股sz Growth大盘Symbol对象
    Growth = Symbol(EMarketTargetType.E_MARKET_TARGET_CN, EMarketSubType.SZ, '399006')
    # a股sh SH300大盘Symbol对象
    SH300 = Symbol(EMarketTargetType.E_MARKET_TARGET_CN, EMarketSubType.SH, '000300')

    # 港股sh大盘HSI对象
    HSI = Symbol(EMarketTargetType.E_MARKET_TARGET_HK, EMarketSubType.HK, 'HSI')
    # 港股sh大盘HSCEI对象
    HSCEI = Symbol(EMarketTargetType.E_MARKET_TARGET_HK, EMarketSubType.HK, 'HSCEI')
    # 港股sh大盘HSCCI对象
    HSCCI = Symbol(EMarketTargetType.E_MARKET_TARGET_HK, EMarketSubType.HK, 'HSCCI')

    # 国内期货只是使用黄金做为时间标尺，不具备对比大盘作用
    BM_FUTURES_CN = Symbol(EMarketTargetType.E_MARKET_TARGET_FUTURES_CN, EMarketSubType.SHFE, 'AU0')
    # 国际期货只是使用纽约黄金做为时间标尺，不具备对比大盘作用
    BM_FUTURES_GB = Symbol(EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL, EMarketSubType.NYMEX, 'GC')

    # 币类只是使用btc做为时间标尺，不具备对比大盘作用
    TC_INX = Symbol(EMarketTargetType.E_MARKET_TARGET_TC, EMarketSubType.COIN, 'btc')



from __future__ import absolute_import

# noinspection PyUnresolvedReferences
from .ABuPickStockExecute import do_pick_stock_work
# noinspection PyUnresolvedReferences
from .ABuPickTimeExecute import do_symbols_with_same_factors, do_symbols_with_diff_factors
# noinspection all
from . import ABuPickTimeWorker as pick_time_worker



# -*- encoding:utf-8 -*-
"""
    预处理函数参数或返回值
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from functools import wraps

from ..CoreBu.ABuFixes import zip
from ..CheckBu.ABuFuncUtil import *

try:
    from ..ExtBu.six.moves import zip_longest
except ImportError:
    from six.moves import zip_longest

__author__ = '夜猫'
__weixin__ = 'abu_quant'


def arg_process(*arg_funcs, **kwarg_funcs):
    """
    【装饰器】
    将funcs函数作用在原函数的参数上；func函数只包括一个参数: return_val.
    :param arg_funcs: func函数tuple。
    :param kwarg_funcs: func函数dict
    :return: 
    """

    def _decorator(func):
        # 检查待bind参数
        check_bind(func, *arg_funcs, **kwarg_funcs)
        # 提取funcs字典
        funcs = bind_partial(func, *arg_funcs, **kwarg_funcs)

        @wraps(func)
        def wrapper(*args, **kwargs):
            # 拼装经过函数处理后的新参数
            args_new = [f(arg) if f else arg for arg, f in zip(args, six.itervalues(funcs))]
            kwargs_new = {k: funcs[k](arg) if k in funcs and funcs[k] else arg
                          for k, arg in six.iteritems(kwargs)}
            # 调用函数
            return func(*args_new, **kwargs_new)

        return wrapper

    return _decorator


def return_process(*funcs, **_unused):
    """
    【装饰器】
    将funcs函数作用在被装饰的函数的返回值上
    :param funcs: 一个或者多个处理对应参数的func函数，多个func函数时需要和返回值数量对齐；func函数只包括一个参数: return_val
    :param _unused: 用来屏蔽dict参数
    :return: 返回被funcs函数处理过的函数返回值
    """
    # 屏蔽dict参数
    if _unused:
        raise TypeError("return_process() doesn't accept dict processors")

    def _decorator(f):
        def wrapper(*args, **kw):
            # 拿到返回值
            return_vals = f(*args, **kw)
            # 用函数处理返回值
            return _apply_funcs(return_vals, funcs)

        return wrapper

    return _decorator


def _apply_funcs(return_vals, funcs):
    """
    将func函数作用在被装饰的函数的返回值上
    """
    # 检查iterable
    if not isinstance(return_vals, tuple):
        return_vals = (return_vals,)
    try:
        iter(funcs)
    except TypeError:
        funcs = (funcs,) if funcs else ()

    # 检查函数
    if not funcs:
        return return_vals

    # 函数和返回值不对齐
    if 1 < len(return_vals) < len(funcs):
        raise TypeError(
            "In _apply_funcs(), len(funcs) == {} more than len(return_vals) == {}".format(
                len(funcs), len(return_vals)
            )
        )
    # 函数多于返回值
    if 1 == len(return_vals) < len(funcs):
        raise TypeError(
            "In _apply_funcs(), only 1 return value with len(processors) == {}".format(len(funcs),
                                                                                       len(return_vals))
        )

    # 将函数作用在返回值上
    return tuple([f(v) if f else v for v, f in zip_longest(return_vals, funcs)])



# -*- encoding:utf-8 -*-
"""
    选股示例因子：相似度选股因子，主要示例fit_first_choice
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np

from ..UtilBu import ABuDateUtil
from ..TLineBu import ABuTLSimilar
from .ABuPickStockBase import AbuPickStockBase, reversed_result
from ..SimilarBu.ABuSimilar import find_similar_with_se, ECoreCorrType

"""外部可通过如：abupy.ps.similar_top = 300修改默认值"""
g_pick_similar_n_top = 100


# noinspection PyAttributeOutsideInit
class AbuPickSimilarNTop(AbuPickStockBase):
    """相似度选股因子示例类"""

    def _init_self(self, **kwargs):
        """通过kwargs设置相似度选股边际条件，相似度计算方法，返回目标数量等，配置因子参数"""

        # 暂时与base保持一致不使用kwargs.pop('a', default)方式
        # 设置目标相似度交易对象
        self.similar_stock = kwargs['similar_stock']

        # 设置目标相似度选取top个数，只在fit_first_choice中使用，默认top100, 即前100个最相似的股票
        self.n_top = g_pick_similar_n_top
        if 'n_top' in kwargs:
            self.n_top = kwargs['n_top']

        # 设置fit_pick中使用的最小相似边际条件
        self.threshold_similar_min = -np.inf
        if 'threshold_similar_min' in kwargs:
            self.threshold_similar_min = kwargs['threshold_similar_min']

        # 设置fit_pick中使用的最大相似边际条件
        self.threshold_similar_max = np.inf
        if 'threshold_similar_max' in kwargs:
            self.threshold_similar_max = kwargs['threshold_similar_max']

        # 相似度是否使用时间加权计算，如果使用速度会慢
        self.rolling = False
        if 'rolling' in kwargs:
            self.rolling = kwargs['rolling']

        # 相似度计算使用算法设置
        self.corr_type = ECoreCorrType.E_CORE_TYPE_PEARS
        if 'corr_type' in kwargs:
            self.corr_type = kwargs['corr_type']
        # 相似度rank缓存
        self.s_sum_rank = None

    @reversed_result
    def fit_pick(self, kl_pd, target_symbol):
        """开始根据自定义相似度边际条件进行选股"""

        # 由于外层worker需要不断迭代symbol使用同一个因子对象进行选股，所以这里缓存了相似度rank结果，只计算一次
        similar_rank, self.s_sum_rank = ABuTLSimilar.calc_similar(self.similar_stock, target_symbol,
                                                                  self.s_sum_rank,
                                                                  show=False)
        # 边际筛选
        if self.threshold_similar_min < similar_rank < self.threshold_similar_max:
            return True
        return False

    def fit_first_choice(self, pick_worker, choice_symbols, *args, **kwargs):
        """
        因子相似度批量选股接口
        :param pick_worker: 选股worker，AbuPickStockWorker实例对象
        :param choice_symbols: 初始备选交易对象序列
        :return:
        """

        # 获取交易目标选股阶段的金融时间序列
        similar_kl_pd = pick_worker.kl_pd_manager.get_pick_stock_kl_pd(self.similar_stock, self.xd, self.min_xd)
        if similar_kl_pd is None or len(similar_kl_pd) == 0:
            return []

        # 选股阶段的金融时间序列最后一个日期作为similar_end
        similar_end = ABuDateUtil.timestamp_to_str(similar_kl_pd.index[-1])
        # 选股阶段的金融时间序列第一个个日期作为similar_start
        similar_start = ABuDateUtil.timestamp_to_str(similar_kl_pd.index[0])
        # 通过ABuSimilar模块中的find_similar_with_se计算与交易目标的相似度rank dict
        net_cg_ret = find_similar_with_se(self.similar_stock, similar_start, similar_end,
                                          rolling=self.rolling, show=False,
                                          corr_type=self.corr_type)
        # 取相似度结果的n_top个，作为选股结果
        similar_top_choice = [ss[0] for ss in net_cg_ret[1:self.n_top + 1]]
        # 通过集合选取在初始备选交易对象序列和相似度选股序列中的子序列
        return list(set(similar_top_choice) & set(choice_symbols))



from __future__ import absolute_import

# noinspection PyUnresolvedReferences
from . import ABuFactorCloseAtrNStop as close
# noinspection PyUnresolvedReferences
from . import ABuFactorPreAtrNStop as pre



# -*- encoding:utf-8 -*-
"""
    字符工具模块
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import random
import re

# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import xrange
from ..CoreBu.ABuFixes import six
from ..CoreBu import ABuEnv

K_CN_RE = re.compile(u'[\u4e00-\u9fa5]+')


def _create_random_tmp(salt_count, seed):
    """
    从seed种子字符池中随机抽取salt_count个字符，返回生成字符串,
    注意抽取属于有放回抽取方法
    :param salt_count: 生成的字符序列的长度
    :param seed: 字符串对象，做为生成序列的种子字符池
    :return: 返回生成字符串
    """
    # TODO random.choice有放回抽取方法, 添加参数支持无放回抽取模式
    sa = [random.choice(seed) for _ in xrange(salt_count)]
    salt = ''.join(sa)
    return salt


def create_random_with_num(salt_count):
    """
    种子字符池 = "0123456789", 从种子字符池中随机抽取salt_count个字符, 返回生成字符串,
    :param salt_count: 生成的字符序列的长度
    :return: 返回生成字符串
    """
    seed = "0123456789"
    return _create_random_tmp(salt_count, seed)


def create_random_with_alpha(salt_count):
    """
    种子字符池 = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
    从种子字符池中随机抽取salt_count个字符, 返回生成字符串,
    :param salt_count: 生成的字符序列的长度
    :return: 返回生成字符串
    """
    seed = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    return _create_random_tmp(salt_count, seed)


def create_random_with_num_low(salt_count):
    """
    种子字符池 = "abcdefghijklmnopqrstuvwxyz0123456789",
    从种子字符池中随机抽取salt_count个字符, 返回生成字符串,
    :param salt_count: 生成的字符序列的长度
    :return: 返回生成字符串
    """
    seed = "abcdefghijklmnopqrstuvwxyz0123456789"
    return _create_random_tmp(salt_count, seed)


def to_unicode(text, encoding=None, errors='strict'):
    """
    to_native_str对py2生效，对six.text_type直接返回，其它的encode，默认utf-8
    """
    if isinstance(text, six.text_type):
        return text
    if not isinstance(text, (bytes, six.text_type)):
        raise TypeError('to_unicode must receive a bytes, str or unicode '
                        'object, got %s' % type(text).__name__)
    if encoding is None:
        encoding = 'utf-8'
    return text.decode(encoding, errors)


def to_bytes(text, encoding=None, errors='strict'):
    """
    to_native_str对py3生效，对bytes直接返回，其它的encode，默认utf-8
    """
    if isinstance(text, bytes):
        return text
    if not isinstance(text, six.string_types):
        raise TypeError('to_bytes must receive a unicode, str or bytes '
                        'object, got %s' % type(text).__name__)
    if encoding is None:
        encoding = 'utf-8'
    return text.encode(encoding, errors)


def to_native_str(text, encoding=None, errors='strict'):
    """
    套接to_unicode和to_bytes针对python版本不同处理

    颠倒转换本来的str类型：
        python2 to_unicode
        python3 to_bytes
    """
    if ABuEnv.g_is_py3:
        return to_bytes(text, encoding, errors)
    else:
        return to_unicode(text, encoding, errors)


def str_is_num10(a_str):
    """通过int(a_str, 10)测试字符串数字是否是10进制"""
    # noinspection PyBroadException
    try:
        int(a_str, 10)
        return True
    except:
        return False


def str_is_num16(a_str):
    """通过int(a_str, 16)测试字符串数字是否是16进制"""
    # noinspection PyBroadException
    try:
        int(a_str, 16)
        return True
    except:
        return False


def str_is_cn(a_str):
    """
        通过正则表达式判断字符串中是否含有中文
        返回结果只判断是否search结果为None, 不返回具体匹配结果
        eg:
            K_CN_RE.search(a_str)('abc') is None
            return False
            K_CN_RE.search(a_str)('abc哈哈') -> <_sre.SRE_Match object; span=(3, 5), match='哈哈'>
            return True
    """
    # a_str = to_unicode(a_str)
    # return any(u'\u4e00' <= c <= u'\u9fa5' for c in a_str)
    return K_CN_RE.search(to_unicode(a_str)) is not None


def digit_str(item):
    """
        从第一个字符开始删除，直到所有字符都是数字为止，或者item长度 < 2
        eg:
            input:  ABuStrUtil.digit_str('sh000001')
            output: 000001

            input:  ABuStrUtil.digit_str('shszsh000001')
            output: 000001
    :param item: 字符串对象
    :return: 过滤head字母的字符串对象
    """
    while True:
        if item.isdigit():
            break
        if len(item) < 2:
            break
        item = item[1:]
    return item


def var_name(var, glb):
    """
    eg：
        in:  a = 5
        in:  var_name(a, globals())
        out: 'a'

    :param var: 要查的变量对象
    :param glb: globals()
    :return: var对象对应的名称
    """
    for vn in glb:
        if glb[vn] is var:
            return vn
    return 'unkonw'



# -*- encoding:utf-8 -*-
"""
    组合参数辅助模块
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import copy
from itertools import product

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from .ABuGridSearch import ParameterGrid
from scipy import stats
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import range, xrange

__author__ = '阿布'
__weixin__ = 'abu_quant'

# TODO 使用enum代替K常量
# 代表买因子参数组合
K_GEN_FACTOR_PARAMS_BUY = 0
# 代表卖因子参数组合
K_GEN_FACTOR_PARAMS_SELL = 1


def gen_factor_grid(type_param, factors):
    """
    :param type_param: grid目标，为K_GEN_FACTOR_PARAMS_BUY或K_GEN_FACTOR_PARAMS_SELL需要重构
    :param factors: 可迭代序列，元素为因子dict 如：
                    {'class': [AbuFactorBuyBreak], 'xd': [42]}, {'class': [AbuFactorBuyBreak],'xd': [60]}
    :return: 返回因子dict的组合参数序列
    """
    
    # 通过ParameterGrid将factor包装，即通过ParameterGrid将dict对象product(*values)，详阅读ParameterGrid
    grid_params = [ParameterGrid(factor) for factor in factors]
    # 进行product调用ParameterGrid__iter__进行product(*values)
    factor_params = product(*grid_params)
    factor_params = [list(pd_cls) for pd_cls in factor_params]

    if len(factors) > 1:
        # 把单独一个factor的加进去
        for grid_single in grid_params:
            for single in grid_single:
                factor_params.append([single])
    if type_param == K_GEN_FACTOR_PARAMS_SELL:
        # 只有sell的factor要加个空的，买的因子要是全空就没办法玩了
        factor_params.append([])  # 最后加一个完全不使用因子的

    return factor_params


def score_pd_plot(grid_score_pd, y_key, x_key=None):
    """对最优结果score可视化，暂时未迁移完整，需迁移其余最优模块后可用"""
    if x_key is not None:
        xt = pd.crosstab(grid_score_pd[x_key], grid_score_pd[y_key])
        xt_pct = xt.div(xt.sum(1).astype(float), axis=0)
        xt_pct.plot(kind='bar',
                    stacked=True,
                    title=str(x_key) + ' -> ' + str(y_key))
        plt.xlabel(str(x_key))
        plt.ylabel(str(y_key))
    else:
        for col in grid_score_pd.columns:
            if col.startswith('Y_'):
                continue
            xt = pd.crosstab(grid_score_pd[col], grid_score_pd[y_key])
            xt_pct = xt.div(xt.sum(1).astype(float), axis=0)
            xt_pct.plot(kind='bar',
                        stacked=True,
                        title=str(col) + ' -> ' + str(y_key))
            plt.xlabel(str(col))
            plt.ylabel(str(y_key))
            plt.show()


# noinspection PyTypeChecker
def make_grid_score_pd(grid_scores, score_index=0):
    """对最优结果score分析处理，暂时未迁移完整，需迁移其余最优模块后可用"""
    unique_sell_factor = {slFac['class'] for grid in grid_scores for slFac in grid[2]}
    unique_buy_factor = {byFac['class'] for grid in grid_scores for byFac in grid[3]}

    grid_pd = pd.DataFrame([alpha_params[1] for alpha_params in grid_scores])
    factor_pd = pd.DataFrame([factor_params[2] for factor_params in grid_scores])
    factor_buy_pd = pd.DataFrame([factor_params[3] for factor_params in grid_scores])

    org_columns = factor_pd.shape[1]
    for columns_ind in xrange(0, org_columns):
        factor_pd[str(columns_ind) + 'class'] = factor_pd[columns_ind].apply(lambda x:
                                                                             None if x is None else x['class'])

    org_buy_columns = factor_buy_pd.shape[1]
    for columns_ind in xrange(0, org_buy_columns):
        factor_buy_pd[str(columns_ind) + 'class'] = factor_buy_pd[columns_ind].apply(lambda x:
                                                                                     None if x is None else x['class'])

    def rm_noise_key(fun_dict):
        copy_dict = copy.deepcopy(fun_dict)
        if 'draw' in copy_dict:
            del copy_dict['draw']
        if 'show' in copy_dict:
            del copy_dict['show']
        if 'class' in copy_dict:
            del copy_dict['class']
        return copy_dict

    def make_factor_pd(x_pd, sel_fac):
        def make_factor_pd_inner(p_x_pd):
            if p_x_pd is None:
                return 0
            if p_x_pd == sel_fac:
                return 1
            return 0

        y_ret = x_pd.apply(make_factor_pd_inner)
        if np.count_nonzero(y_ret) > 0:
            return 1
        return 0

    def make_factor_param_pd(x_pd, unique_facts):
        def make_factor_param_pd_inner(p_x_pd):
            if p_x_pd is None:
                return
            class_key = p_x_pd['class'].__name__
            x_pd_copy = rm_noise_key(p_x_pd)
            for item in x_pd_copy.items():
                unique_item = class_key + ':' + str(item[0]) + ':' + str(item[1])
                unique_facts.append(unique_item)

        x_pd.apply(make_factor_param_pd_inner)

    for selFac in unique_sell_factor:
        grid_pd[selFac.__name__] = factor_pd.iloc[:, org_columns:].apply(
            make_factor_pd, args=(selFac,), axis=1)
    for buyFac in unique_buy_factor:
        grid_pd[buyFac.__name__] = factor_buy_pd.iloc[:, org_buy_columns:].apply(
            make_factor_pd, args=(buyFac,), axis=1)

    unique_factor_params = []
    unique_factor_buy_params = []
    factor_pd.iloc[:, 0:org_columns].apply(make_factor_param_pd, args=(unique_factor_params,), axis=1)
    factor_buy_pd.iloc[:, 0:org_buy_columns].apply(make_factor_param_pd, args=(unique_factor_buy_params,), axis=1)

    unique_factor_params = set(unique_factor_params)
    unique_factor_buy_params = set(unique_factor_buy_params)

    def dummies_params(x_pd, p_sel_fac_param):
        def dummies_params_inner(p_x_pd):
            if p_x_pd is None:
                return 0
            class_key = p_x_pd['class'].__name__
            x_pd_copy = rm_noise_key(p_x_pd)
            for item in x_pd_copy.items():
                unique_item = class_key + ':' + str(item[0]) + ':' + str(item[1])
                if p_sel_fac_param == unique_item:
                    return 1
            return 0

        y_ret = x_pd.apply(dummies_params_inner)
        if np.count_nonzero(y_ret) > 0:
            return 1
        return 0

    for sel_fac_param in unique_factor_params:
        grid_pd[sel_fac_param] = factor_pd.iloc[:, 0:org_columns].apply(
            dummies_params, args=(sel_fac_param,), axis=1)
    for buy_fac_param in unique_factor_buy_params:
        grid_pd[buy_fac_param] = factor_pd.iloc[:, 0:org_columns].apply(
            dummies_params, args=(buy_fac_param,), axis=1)

    grid_pd['Y_REG'] = [score[0][score_index] if isinstance(score[0], list) else score[0]
                        for score in grid_scores]

    grid_pd['Y_LOG_MEDIAN'] = np.where(
        grid_pd['Y_REG'] > grid_pd['Y_REG'].median(), 1, 0)
    grid_pd['Y_LOG_618'] = np.where(
        grid_pd['Y_REG'] > stats.scoreatpercentile(grid_pd['Y_REG'], 61.8), 1, 0)

    return grid_pd



# -*- encoding:utf-8 -*-
"""期货度量模块"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from ..CoreBu import ABuEnv
from ..ExtBu.empyrical import stats
from ..MetricsBu.ABuMetricsBase import AbuMetricsBase, valid_check
from ..UtilBu.ABuDTUtil import warnings_filter

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuMetricsFutures(AbuMetricsBase):
    """期货度量类，主要区别在于不涉及benchmark"""

    def _metrics_base_stats(self):
        """度量真实成交了的capital_pd，即涉及资金的度量，期货相关不涉及benchmark"""
        # 平均资金利用率
        self.cash_utilization = 1 - (self.capital.capital_pd.cash_blance /
                                     self.capital.capital_pd.capital_blance).mean()
        self.algorithm_returns = np.round(self.capital.capital_pd['capital_blance'].pct_change(), 3)

        # 收益cum数据
        # noinspection PyTypeChecker
        self.algorithm_cum_returns = stats.cum_returns(self.algorithm_returns)

        # 最后一日的cum return
        self.algorithm_period_returns = self.algorithm_cum_returns[-1]

        # 交易天数
        self.num_trading_days = len(self.algorithm_cum_returns)

        # 年化收益
        self.algorithm_annualized_returns = \
            (ABuEnv.g_market_trade_year / self.num_trading_days) * self.algorithm_period_returns

        # noinspection PyUnresolvedReferences
        self.mean_algorithm_returns = self.algorithm_returns.cumsum() / np.arange(1, self.num_trading_days + 1,
                                                                                  dtype=np.float64)
        # noinspection PyTypeChecker
        self.algorithm_volatility = stats.annual_volatility(self.algorithm_returns)
        # noinspection PyTypeChecker
        self.algorithm_sharpe = stats.sharpe_ratio(self.algorithm_returns)
        # 最大回撤
        # noinspection PyUnresolvedReferences
        self.max_drawdown = stats.max_drawdown(self.algorithm_returns.values)

    @valid_check
    @warnings_filter  # skip: statsmodels / nonparametric / kdetools.py:20
    def plot_returns_cmp(self, only_show_returns=False, only_info=False):
        """考虑资金情况下的度量，进行与benchmark的收益度量对比，收益趋势，资金变动可视化，以及其它度量信息，不涉及benchmark"""

        self.log_func('买入后卖出的交易数量:{}'.format(self.order_has_ret.shape[0]))
        self.log_func('胜率:{:.4f}%'.format(self.win_rate * 100))

        self.log_func('平均获利期望:{:.4f}%'.format(self.gains_mean * 100))
        self.log_func('平均亏损期望:{:.4f}%'.format(self.losses_mean * 100))

        self.log_func('盈亏比:{:.4f}'.format(self.win_loss_profit_rate))

        self.log_func('策略收益: {:.4f}%'.format(self.algorithm_period_returns * 100))
        self.log_func('策略年化收益: {:.4f}%'.format(self.algorithm_annualized_returns * 100))

        self.log_func('策略买入成交比例:{:.4f}%'.format(self.buy_deal_rate * 100))
        self.log_func('策略资金利用率比例:{:.4f}%'.format(self.cash_utilization * 100))
        self.log_func('策略共执行{}个交易日'.format(self.num_trading_days))

        if only_info:
            return

        self.algorithm_cum_returns.plot()
        plt.legend(['algorithm returns'], loc='best')
        plt.show()

        if only_show_returns:
            return
        sns.regplot(x=np.arange(0, len(self.algorithm_cum_returns)), y=self.algorithm_cum_returns.values)
        plt.show()
        sns.distplot(self.capital.capital_pd['capital_blance'], kde_kws={"lw": 3, "label": "capital blance kde"})
        plt.show()

    @valid_check
    def plot_sharp_volatility_cmp(self, only_info=False):
        """sharp，volatility信息输出"""

        self.log_func('策略Sharpe夏普比率: {:.4f}'.format(self.algorithm_sharpe))
        self.log_func('策略波动率Volatility: {:.4f}'.format(self.algorithm_volatility))



# -*- encoding:utf-8 -*-
"""
    通用装饰器工具模块
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import logging
import pdb
import time
import warnings
from collections import Iterable

import numpy as np
import pandas as pd

from ..CoreBu import ABuEnv
from ..CoreBu.ABuFixes import six


def warnings_filter(func):
    """
        作用范围：函数装饰器 (模块函数或者类函数)
        功能：被装饰的函数上的警告不会打印，忽略
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        warnings.simplefilter('ignore')
        ret = func(*args, **kwargs)
        if not ABuEnv.g_ignore_all_warnings:
            # 如果env中的设置不是忽略所有才恢复
            warnings.simplefilter('default')
        return ret

    return wrapper


def singleton(cls):
    """
        作用范围：类装饰器
        功能：被装饰后类变成单例类
    """

    instances = {}

    @functools.wraps(cls)
    def get_instance(*args, **kw):
        if cls not in instances:
            # 不存在实例instances才进行构造
            instances[cls] = cls(*args, **kw)
        return instances[cls]

    return get_instance


# TODO 放在这里不合适，还要和ABuScalerUtil中的装饰器arr_to_pandas重复代码进行重构
def arr_to_pandas(arr):
    """
        函数装饰器：将可以迭代的序列转换为pd.DataFrame或者pd.Series，支持
        np.ndarray，list，dict, list，set，嵌套可迭代序列, 混嵌套可迭代序列
    """
    # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
    if not isinstance(arr, Iterable) or isinstance(arr, six.string_types):
        return arr

    if not isinstance(arr, pd.DataFrame) or isinstance(arr, pd.Series):
        if isinstance(arr, np.ndarray) and len(arr.shape) > 1 and arr.shape[1] > 1:
            # np.ndarray > 1列的转换为pd.DataFrame
            arr = pd.DataFrame(arr)
        elif isinstance(arr, dict):
            # 针对dict转换pd.DataFrame，注意顺序不能在all(isinstance(arr_item, Iterable)下面
            arr = pd.DataFrame(arr)
        elif all(isinstance(arr_item, Iterable) for arr_item in arr):
            # 如果子序列的元素也都是可以迭代的，那么先转np.array，然后再DataFrame
            arr = pd.DataFrame(np.array(arr))
        else:
            # 否则序列对象转换为pd.Series
            arr = pd.Series(arr)
    return arr


def params_to_pandas(func):
    """
        函数装饰器：不定参数装饰器，定参数转换使用ABuScalerUtil中的装饰器arr_to_pandas(func)
        将被装饰函数中的参数中所有可以迭代的序列转换为pd.DataFrame或者pd.Series
    """

    @functools.wraps(func)
    def wrapper(*arg, **kwargs):
        # 把arg中的可迭代序列转换为pd.DataFrame或者pd.Series
        arg_list = [arr_to_pandas(param) for param in arg]
        # 把kwargs中的可迭代序列转换为pd.DataFrame或者pd.Series
        arg_dict = {param_key: arr_to_pandas(kwargs[param_key]) for param_key in kwargs}
        return func(*arg_list, **arg_dict)

    return wrapper


# TODO 放在这里不合适，还要和ABuScalerUtil中的装饰器arr_to_numpy重复代码进行重构
def arr_to_numpy(arr):
    """
        函数装饰器：将可以迭代的序列转换为np.array，支持pd.DataFrame或者pd.Series
        ，list，dict, list，set，嵌套可迭代序列, 混嵌套可迭代序列
    """
    # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
    if not isinstance(arr, Iterable) or isinstance(arr, six.string_types):
        return arr

    if not isinstance(arr, np.ndarray):
        if isinstance(arr, pd.DataFrame) or isinstance(arr, pd.Series):
            # 如果是pandas直接拿values
            arr = arr.values
        elif isinstance(arr, dict):
            # 针对dict转换np.array
            arr = np.array(list(arr.values())).T
        else:
            arr = np.array(arr)
    return arr


def params_to_numpy(func):
    """
        函数装饰器：不定参数装饰器，定参数转换使用ABuScalerUtil中的装饰器arr_to_numpy(func)
        将被装饰函数中的参数中所有可以迭代的序列转换为np.array
    """

    @functools.wraps(func)
    def wrapper(*arg, **kwargs):
        # 把arg中的可迭代序列转换为np.array
        arg_list = [arr_to_numpy(param) for param in arg]
        # 把kwargs中的可迭代序列转换为np.array
        arg_dict = {param_key: arr_to_numpy(kwargs[param_key]) for param_key in kwargs}
        return func(*arg_list, **arg_dict)

    return wrapper


def catch_error(return_val=None, log=True):
    """
    作用范围：函数装饰器 (模块函数或者类函数)
    功能：捕获被装饰的函数中所有异常，即忽略函数中所有的问题，用在函数的执行级别低，且不需要后续处理
    :param return_val: 异常后返回的值，
                eg:
                    class A:
                        @ABuDTUtil.catch_error(return_val=100)
                        def a_func(self):
                            raise ValueError('catch_error')
                            return 100
                    in: A().a_func()
                    out: 100
    :param log: 是否打印错误日志
    """
    def decorate(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                logging.exception(e) if log else logging.debug(e)
                return return_val

        return wrapper

    return decorate


def consume_time(func):
    """
    作用范围：函数装饰器 (模块函数或者类函数)
    功能：简单统计被装饰函数运行时间
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        print('{} cost {}s'.format(func.__name__, round(end_time - start_time, 3)))
        return result

    return wrapper


def empty_wrapper(func):
    """
    作用范围：函数装饰器 (模块函数或者类函数)
    功能：空装饰器，为fix版本问题使用，或者分逻辑功能实现使用
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


# noinspection PyUnusedLocal
def empty_wrapper_with_params(*p_args, **p_kwargs):
    """
    作用范围：函数装饰器 (模块函数或者类函数)
    功能：带参数空装饰器，为fix版本问题使用，或者分逻辑功能实现使用
    """

    def decorate(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return decorate


def except_debug(func):
    """
    作用范围：函数装饰器 (模块函数或者类函数)
    功能：debug，调试使用，装饰在有问题函数上，发生问题打出问题后，再运行一次函数，可以用s跟踪问题了
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            pdb.set_trace()
            print(e)
            # 再来一遍用s跟踪进去
            return func(*args, **kwargs)

    return wrapper



# -*- encoding:utf-8 -*-
"""
    技术指标工具基础模块
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import logging

import pandas as pd
from enum import Enum

from ..UtilBu import ABuDateUtil

__author__ = '阿布'
__weixin__ = 'abu_quant'


class ECalcType(Enum):
    """
        技术指标技术方式类
    """
    """使用talib透传技术指标计算"""
    E_FROM_TA = 0
    """使用pandas等库实现技术指标计算"""
    E_FROM_PD = 1


# try:
#     # 不强制要求talib，全部局部引用
#     # noinspection PyUnresolvedReferences
#     import talib
#     g_calc_type = ECalcType.E_FROM_TA
# except ImportError:
#     # 没有安装talib，使用E_FROM_PD
#     g_calc_type = ECalcType.E_FROM_PD
"""彻底不用talib，完全都使用自己计算的指标结果"""
g_calc_type = ECalcType.E_FROM_PD


def plot_from_order(plot_nd_func, order, date_ext, **kwargs):
    """
    封装在技术指标上绘制交易order信号通用流程
    :param plot_nd_func: 绘制技术指标的具体实现函数，必须callable
    :param order: AbuOrder对象转换的pd.DataFrame对象or pd.Series对象
    :param date_ext: int对象 eg. 如交易在2015-06-01执行，如date_ext＝120，择start向前推120天，end向后推120天
    :param kwargs: plot_nd_func需要的其它关键字参数，直接透传给plot_nd_func
    """
    if not callable(plot_nd_func):
        # plot_nd_func必须是callable
        raise TypeError('plot_nd_func must callable!!')

    if not isinstance(order, (pd.DataFrame, pd.Series)) and order.shape[0] > 0:
        # order必须是pd.DataFrame对象or pd.Series对象 且 单子数量要 > 0
        raise TypeError('order must DataFrame here!!')

    is_df = isinstance(order, pd.DataFrame)

    if is_df and order.shape[0] == 1:
        # 如果是只有1行pd.DataFrame对象则变成pd.Series
        is_df = False
        # 通过iloc即变成pd.Series对象
        # noinspection PyUnresolvedReferences
        order = order.iloc[0]

    def plot_from_series(p_order):
        """
        根据交易的symbol信息买入，卖出时间，以及date_ext完成通过ABuSymbolPd.make_kl_df获取金融时间序列，
        在成功获取数据后使用plot_nd_func完成买入卖出信号绘制及对应的技术指标绘制
        :param p_order: AbuOrder对象转换的pd.Series对象
        """
        # 确定交易对象
        target_symbol = p_order['symbol']
        # 单子都必须有买入时间
        buy_index = pd.to_datetime(str(p_order['buy_date']))
        sell_index = None

        start = ABuDateUtil.fmt_date(p_order['buy_date'])
        # 通过date_ext确定start，即买人单子向前推date_ext天
        start = ABuDateUtil.begin_date(date_ext, date_str=start, fix=False)
        if p_order['sell_type'] != 'keep':
            sell_index = pd.to_datetime(str(p_order['sell_date']))
            # 如果有卖出，继续通过sell_date，date_ext确定end时间
            end = ABuDateUtil.fmt_date(p_order['sell_date'])
            # -date_ext 向前
            end = ABuDateUtil.begin_date(-date_ext, date_str=end, fix=False)
        else:
            end = None
        from ..MarketBu import ABuSymbolPd
        # 组织好参数，确定了请求范围后开始获取金融时间序列数据
        kl_pd = ABuSymbolPd.make_kl_df(target_symbol, start=start, end=end)
        if kl_pd is None or kl_pd.shape[0] == 0:
            logging.debug(target_symbol + ': has net error in data')
            return
        # 使用plot_nd_func完成买入卖出信号绘制及对应的技术指标绘制
        return plot_nd_func(kl_pd, with_points=buy_index, with_points_ext=sell_index, **kwargs)

    if not is_df:
        return plot_from_series(order)
    else:
        # 多个order, apply迭代执行plot_from_series
        order = order[order['result'] != 0]
        return order.apply(plot_from_series, axis=1)



# -*- encoding:utf-8 -*-
"""
    相关系数相似应用模块
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import functools
import logging
import math
import operator
import os

import pandas as pd

from . import ABuCorrcoef
from . import ABuSimilarDrawing
from .ABuCorrcoef import ECoreCorrType
from ..TradeBu import AbuBenchmark
from ..CoreBu import ABuEnv
from ..CoreBu.ABuParallel import delayed, Parallel
from ..CoreBu.ABuEnv import EMarketDataSplitMode, EMarketTargetType
from ..MarketBu import ABuSymbolPd
from ..MarketBu.ABuMarket import split_k_market, all_symbol
from ..MarketBu.ABuSymbol import IndexSymbol
from ..UtilBu.ABuDTUtil import consume_time
from ..UtilBu.ABuProgress import do_clear_output
from ..CoreBu.ABuEnvProcess import add_process_env_sig, AbuEnvProcess
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import xrange
from ..UtilBu import ABuProgress

"""进行相似度数据收集并行进程数，IO操作偏多，所以分配多个，默认=cpu个数＊2, windows还是..."""
g_process_panel_cnt = ABuEnv.g_cpu_cnt * 2 if ABuEnv.g_is_mac_os else ABuEnv.g_cpu_cnt


def from_local(func):
    """
    现所有相似度应用默认为from_local模式，即需要在有数据的情况下做相似度应用

    为进行相似度数据收集的函数装饰，作用是忽略env中的数据获取模式，改变数据获取模式，
    只使用本地数据模式进行数据收集，完成整个任务后，再恢复之前的数据获取模式
    :param func: 进行相似度应用且有数据收集行为的函数
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # 临时保存env设置中的g_data_fetch_mode
        fetch_mode = ABuEnv.g_data_fetch_mode
        # 设置数据获取模式为强制本地缓存模式
        ABuEnv.g_data_fetch_mode = ABuEnv.EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL
        if fetch_mode != ABuEnv.EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL:
            # 如果原有设置不是强制本地缓存模式，warning提示
            logging.warning('data from local. run ABu.run_kl_update if you want to get the latest data.')
        result = func(*args, **kwargs)
        # 恢复之前的g_data_fetch_mode
        ABuEnv.g_data_fetch_mode = fetch_mode
        return result

    return wrapper


def from_net(func):
    """
    为进行相似度数据收集的函数装饰，作用是忽略env中的数据获取模式，改变数据获取模式，
    只使用网络数据模式进行数据收集，完成整个任务后，再恢复之前的数据获取模式
    :param func: 进行相似度应用且有数据收集行为的函数
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # 临时保存env设置中的g_data_fetch_mode
        fetch_mode = ABuEnv.g_data_fetch_mode
        # 设置数据获取模式为强制网络模式
        ABuEnv.g_data_fetch_mode = ABuEnv.EMarketDataFetchMode.E_DATA_FETCH_FORCE_NET
        if fetch_mode != ABuEnv.EMarketDataFetchMode.E_DATA_FETCH_FORCE_NET:
            # 如果原有设置不是强制网络模式，warning提示
            logging.warning('data from net!!!')
        result = func(*args, **kwargs)
        # 恢复之前的g_data_fetch_mode
        ABuEnv.g_data_fetch_mode = fetch_mode
        return result

    return wrapper


@from_local
def _find_similar(symbol, cmp_cnt=None, n_folds=None, start=None, end=None, show_cnt=None, rolling=False,
                  show=True, corr_type=ECoreCorrType.E_CORE_TYPE_PEARS):
    """
    被from_local装饰器装饰 即强制走本地数据，获取全市场symbol涨跌幅度pd.DataFrame对象，
    使用symbol涨跌幅度与全市场symbol涨跌幅度进行相关对比，可视化结果及信息
    :param symbol: 外部指定目标symbol，str对象
    :param cmp_cnt: 相关性对比多少个交易日，int，可选参数
    :param n_folds: 相关性对比n_folds年，int，可选参数
    :param start: 请求的开始日期str对象，可选参数
    :param end: 请求的结束日期str对象，可选参数
    :param show_cnt: 最终结果展示以及可视化相似度个数
    :param rolling: 是否使用时间加权相关计算，与corr_type=ECoreCorrType.E_CORE_TYPE_ROLLING一样，单独拿出来了
    :param show: 是否可视化最终top最相关的股票
    :param corr_type: ECoreCorrType对象，暂时支持皮尔逊，斯皮尔曼，＋－符号相关系数，移动时间加权相关系数
    """
    # 获取全市场symbol涨跌幅度pd.DataFrame对象
    market_change_df = _all_market_cg(symbol, cmp_cnt=cmp_cnt, n_folds=n_folds, start=start, end=end)
    if market_change_df is None:
        logging.info('{} data is miss, please update data first!'.format(symbol))
        return
    # 重新赋予标尺实际的交易日数量
    cmp_cnt = market_change_df[symbol].shape[0]
    # symbol涨跌幅度df数据
    benchmark_df = market_change_df[symbol]
    # 清一下输出，太乱
    do_clear_output()
    # 开始使用symbol涨跌幅度与全市场symbol涨跌幅度进行相关对比，可视化结果及信息
    sorted_corr = _handle_market_change_df(market_change_df, cmp_cnt, benchmark_df, show_cnt,
                                           corr_type, rolling, show)
    return sorted_corr


def find_similar_with_se(symbol, start, end, show_cnt=10, rolling=False, show=True,
                         corr_type=ECoreCorrType.E_CORE_TYPE_PEARS):
    """
    固定参数使用start, end参数提供时间范围规则，套接_find_similar，为_find_similar提供时间范围规则
    :param symbol: 外部指定目标symbol，str对象
    :param start: 请求的开始日期str对象
    :param end: 请求的结束日期str对象
    :param show_cnt: 最终结果展示以及可视化相似度个数
    :param rolling: 是否使用时间加权相关计算，与corr_type=ECoreCorrType.E_CORE_TYPE_ROLLING一样，单独拿出来了
    :param show: 是否可视化最终top最相关的股票
    :param corr_type: ECoreCorrType对象，暂时支持皮尔逊，斯皮尔曼，＋－符号相关系数，移动时间加权相关系数
    :return:
    """
    return _find_similar(symbol, start=start, end=end, show_cnt=show_cnt, rolling=rolling, show=show,
                         corr_type=corr_type)


def find_similar_with_folds(symbol, n_folds=None, show_cnt=10, rolling=False, show=True,
                            corr_type=ECoreCorrType.E_CORE_TYPE_PEARS):
    """
    固定参数使用n_folds参数提供时间范围规则，套接_find_similar，为_find_similar提供时间范围规则
    :param symbol: 外部指定目标symbol，str对象
    :param n_folds: 相关性对比n_folds年，int
    :param show_cnt: 最终结果展示以及可视化相似度个数
    :param rolling: 是否使用时间加权相关计算，与corr_type=ECoreCorrType.E_CORE_TYPE_ROLLING一样，单独拿出来了
    :param show: 是否可视化最终top最相关的股票
    :param corr_type: ECoreCorrType对象，暂时支持皮尔逊，斯皮尔曼，＋－符号相关系数，移动时间加权相关系数
    :return:
    """
    return _find_similar(symbol, n_folds=n_folds, show_cnt=show_cnt, rolling=rolling, show=show,
                         corr_type=corr_type)


def find_similar_with_cnt(symbol, cmp_cnt=60, show_cnt=10, rolling=False, show=True,
                          corr_type=ECoreCorrType.E_CORE_TYPE_PEARS):
    """
    固定参数使用cmp_cnt参数提供时间范围规则，套接_find_similar，为_find_similar提供时间范围规则
    :param symbol: 外部指定目标symbol，str对象
    :param cmp_cnt: 相关性对比多少个交易日，int
    :param show_cnt: 最终结果展示以及可视化相似度个数
    :param rolling: 是否使用时间加权相关计算，与corr_type=ECoreCorrType.E_CORE_TYPE_ROLLING一样，单独拿出来了
    :param show: 是否可视化最终top最相关的股票
    :param corr_type: ECoreCorrType对象，暂时支持皮尔逊，斯皮尔曼，＋－符号相关系数，移动时间加权相关系数
    :return:
    """
    return _find_similar(symbol, cmp_cnt=cmp_cnt, show_cnt=show_cnt, rolling=rolling, show=show,
                         corr_type=corr_type)


@add_process_env_sig
def _make_symbols_cg_df(symbols, benchmark):
    """
    相关性金融数据收集，子进程委托函数，子进程通过make_kl_df完成主进程委托的symbols个
    金融数据收集工作，最终返回所有金融时间序列涨跌幅度pd.DataFrame对象
    :param symbols: 可迭代symbols序列，序列中的元素为str对象
    :param benchmark: 进行数据收集使用的标尺对象，数据时间范围确定使用，AbuBenchmark实例对象
    :return: 所有金融时间序列涨跌幅度pd.DataFrame对象
    """

    # 子进程金融数据收集工作, 由于本事是在子进程内工作，所以不再make_kl_df中使用parallel模式，上层进行多任务分配及任务数确定
    panel = ABuSymbolPd.make_kl_df(symbols, data_mode=EMarketDataSplitMode.E_DATA_SPLIT_UNDO, benchmark=benchmark,
                                   show_progress=True)

    if panel is None or panel.empty:
        logging.info('pid {} panel is None'.format(os.getpid()))
        return None
    # 转换panel轴方向，即可方便获取所有金融时间数据的某一个列
    panel = panel.swapaxes('items', 'minor')
    net_cg_df = panel['p_change'].fillna(value=0)
    """
        转轴后直接获取p_change，即所有金融时间序列涨跌幅度pd.DataFrame对象，形如下所示：
                    usF	    usFCAU	usGM	usHMC	usTM	usTSLA	usTTM
        2015-06-25	-0.387	-0.517	-1.308	0.522	-0.391	1.365	-0.029
        2015-06-26	-0.259	1.300	-0.922	0.366	0.437	-0.632	-0.229
        2015-06-29	-2.468	-6.799	-3.345	-2.676	-2.222	-1.898	-2.550
        2015-06-30	-0.067	0.000	0.301	1.250	0.982	2.381	1.353
        2015-07-01	-0.133	0.688	-0.870	-1.605	-0.112	0.332	0.261
        .................................................................
    """
    return net_cg_df


def _make_benchmark_cg_df(symbol, benchmark):
    """
    根据benchmark提取其时间序列对象kl_pd中的p_change列，返回p_change组成的
    新pd.DataFrame对象，行序列名即为symbol
    :param symbol: 标尺对象symbol，str对象
    :param benchmark: 进行数据收集使用的标尺对象，数据时间范围确定使用，AbuBenchmark实例对象
    :return: 返回p_change组成的新pd.DataFrame对象
    """
    kl_pd = benchmark.kl_pd
    net_cg_df = pd.DataFrame({symbol: kl_pd['p_change']}, index=kl_pd.index).fillna(value=0)
    """
        p_change组成的新pd.DataFrame对象，行序列名即为symbol, 形如下所示：
                    us.IXIC
        2014-07-25	-0.50
        2014-07-28	-0.10
        2014-07-29	-0.05
        2014-07-30	0.45
        2014-07-31	-2.09
        2014-08-01	-0.39
        2014-08-04	0.72
    """
    return net_cg_df


def _net_cg_df_create(symbol, benchmark):
    """
    获取env中全市场symbol，切分分配子进程，委托子进程_make_symbols_cg_df函数，
    将子进程返回的金融时间序列涨跌幅度pd.DataFrame对象再次进行连接，组合成为全市场
    symbol涨跌幅度pd.DataFrame对象
    :param symbol: 标尺对象symbol，str对象
    :param benchmark: 进行数据收集使用的标尺对象，数据时间范围确定使用，AbuBenchmark实例对象
    :return: 全市场symbol涨跌幅度pd.DataFrame对象
    """

    # 获取全市场symbol，没有指定市场参数，即根据env中设置的市场来获取所有市场symbol
    choice_symbols = all_symbol()
    # 通过split_k_market将市场symbol切割为子进程需要完成的任务数量
    process_symbols = split_k_market(g_process_panel_cnt, market_symbols=choice_symbols)
    # 因为切割会有余数，所以将原始设置的进程数切换为分割好的个数, 即32 -> 33 16 -> 17
    n_process_pick_stock = len(process_symbols)
    parallel = Parallel(
        n_jobs=n_process_pick_stock, verbose=0, pre_dispatch='2*n_jobs')

    # 暂时关闭多进程进度条，太多, 注意这种全局设置一定要在AbuEnvProcess初始化之前完成
    # ABuProgress.g_show_ui_progress = False
    # _make_symbols_cg_df被装饰器add_process_env_sig装饰，需要进程间内存拷贝对象AbuEnvProcess, 详AbuEnvProcess
    p_nev = AbuEnvProcess()
    change_df_array = parallel(
        delayed(_make_symbols_cg_df)(choice_symbols, benchmark, env=p_nev) for choice_symbols in process_symbols)
    # ABuProgress.g_show_ui_progress = True
    # 还是显示进度条，但是完事时检测一下还有没有ui进度条
    ABuProgress.do_check_process_is_dead()
    """
        如果标尺的涨跌幅已经在choice_symbols中就不单独获取组装了，没有的情况是如：
        eg. env中指定市场参数港股，即g_market_target = EMarketTargetType.E_MARKET_TARGET_HK，但是
        传人的symbol是a股市场中的一支股票，即目的是想从整个港股市场中分析与这支a股股票的相关系数，这时即会
        触发_make_benchmark_cg_df的使用
    """
    change_df_concat = None if symbol in choice_symbols else _make_benchmark_cg_df(symbol, benchmark)
    for change_df in change_df_array:
        if change_df is not None:
            # 将所有子进程返回的金融时间序列涨跌幅度pd.DataFrame对象再次进行连接
            change_df_concat = change_df if change_df_concat is None else pd.concat([change_df, change_df_concat],
                                                                                    axis=1)
    return change_df_concat


@consume_time
def _all_market_cg(symbol, cmp_cnt=None, n_folds=None, start=None, end=None):
    """
    获取全市场symbol涨跌幅度pd.DataFrame对象
    :param symbol: 外部指定目标symbol，str对象
    :param cmp_cnt: 对比多少个交易日，int，可选参数
    :param n_folds: 对比n_folds年，int，可选参数
    :param start: 请求的开始日期 str对象，可选参数
    :param end: 请求的结束日期 str对象，可选参数
    :return: 全市场symbol涨跌幅度pd.DataFrame对象, 形如下所示：
                        e.g.
                            usA	    usAA	usAAC
                2015/7/27	0.76	-1.94	0.59
                2015/7/28	2.12	2.6	    1.3
                2015/7/29	-0.12	2.94	-1.34
                2015/7/30	1.41	-1.77	-4.04
                2015/7/31	-0.05	-1.1	1.39
    """

    if cmp_cnt is not None:
        # 如果有传递对比多少个交易日这个参数，即反向修改n_folds，ceil向上对齐金融序列获取年数
        n_folds = int(math.ceil(cmp_cnt / ABuEnv.g_market_trade_year))
    # 标尺不是使用大盘symbol，而是传人的symbol做为标尺
    benchmark = AbuBenchmark(benchmark=symbol, n_folds=n_folds, start=start, end=end, rs=False)
    if benchmark.kl_pd is None or benchmark.kl_pd.empty:
        logging.info('{} make benchmark get None'.format(symbol))
        return None

    if cmp_cnt is not None and benchmark.kl_pd.shape[0] > cmp_cnt:
        # 再次根据对比多少个交易日这个参数，对齐时间序列
        benchmark.kl_pd = benchmark.kl_pd.iloc[-cmp_cnt:]
    # 有了symbol和benchmark，即可开始获取全市场symbol涨跌幅度pd.DataFrame对象all_market_change_df
    all_market_change_df = _net_cg_df_create(symbol, benchmark)
    return all_market_change_df


def _handle_market_change_df(market_change_df, cmp_cnt, benchmark_df, show_cnt, corr_type, rolling=True, show=True):
    """
    使用benchmark_df与全市场market_change_df进行相关系数计算，可视化结果及信息
    :param market_change_df: 全市场symbol涨跌幅度pd.DataFrame对象
    :param cmp_cnt: 对比多少个交易日，int
    :param benchmark_df: 标尺symbol对应的pd.Series对象
    :param show_cnt: 最终结果展示以及可视化相似度个数
    :param corr_type: ECoreCorrType对象，暂时支持皮尔逊，斯皮尔曼，＋－符号相关系数，移动时间加权相关系数
    :param rolling: 是否使用时间加权相关计算，与corr_type = ECoreCorrType.E_CORE_TYPE_ROLLING一样，单独拿出来了
    :param show: 是否可视化最终top最相关的股票
    :return:
    """
    # 使用[-cmp_cnt:]再次确定时间序列周期
    benchmark_df = benchmark_df.iloc[-cmp_cnt:]
    market_change_df = market_change_df.iloc[-cmp_cnt:]

    if corr_type == ECoreCorrType.E_CORE_TYPE_ROLLING:
        # 把参数时间加权rolling和corr_type设置进行merge
        rolling = True

    if rolling:
        # 时间加权统一使用ABuCorrcoef.rolling_corr单独计算，即使用两个参数方式计算，详见ABuCorrcoef.rolling_corr
        corr_ret = ABuCorrcoef.rolling_corr(market_change_df, benchmark_df)
        corr_ret = pd.Series(corr_ret, index=market_change_df.columns, name=benchmark_df.name)
    else:
        # 其它加权计算统一使用corr_df计算，即统一使用大矩阵计算相关系数后再拿出benchmark_df对应的相关系数列
        corr_ret = ABuCorrcoef.corr_matrix(market_change_df, corr_type)[benchmark_df.name]
    # 对结果进行zip排序，按照相关系统由正相关到负相关排序
    sorted_ret = sorted(zip(corr_ret.index, corr_ret), key=operator.itemgetter(1), reverse=True)
    """
        最终sorted_ret为可迭代序列，形如：
        [('usTSLA', 1.0), ('usSINA', 0.45565379371028253), ('usWB', 0.44811939073120288),
         ('usAEH', 0.37792534372729375), ('usCRESY', 0.37347584342214574),
         ('us.IXIC', 0.36856818073255937), ('usCVG', 0.36841463066151853),
         ('usOCN', 0.36412381487296047), ('usYHOO', 0.36217456000137549), ...............]
    """
    if show:
        # 根据是否是ipython环境决定信息输出函数
        log_func = logging.info if ABuEnv.g_is_ipython else print
        log_func(sorted_ret[:show_cnt])
        # 绘制show_cnt个最相关的股票股价走势
        ABuSimilarDrawing.draw_show_close(sorted_ret, cmp_cnt, show_cnt)

    return sorted_ret


@consume_time
@from_local
def multi_corr_df(corr_jobs, cmp_cnt=252, n_folds=None, start=None, end=None):
    """
    被from_local装饰器装饰 即强制走本地数据，匹配市场对应的benchmark，根据参数
    使用_all_market_cg获取全市场symbol涨跌幅度pd.DataFrame对象change_df使用
    corr_jobs个相关系数计算方法分别计算change_df的相关系数，所有结果组成一个字典返回
    :param corr_jobs: 需要执行相关计算方法ECoreCorrType序列
    :param cmp_cnt: 对比多少个交易日，int
    :param n_folds: 对比n_folds年，int，可选参数
    :param start: 请求的开始日期 str对象，可选参数
    :param end: 请求的结束日期 str对象，可选参数
    :return: 返回相关系数矩阵组成的字典对象，如下所示 eg：

            {'pears':
                            usBIDU    usFB  usGOOG  usNOAH  usSFUN  usTSLA  usVIPS  usWUBA
            usBIDU         1.0000  0.3013  0.3690  0.4015  0.3680  0.3015  0.3706  0.4320
            usFB           0.3013  1.0000  0.6609  0.2746  0.1978  0.4080  0.2856  0.2438
            usGOOG         0.3690  0.6609  1.0000  0.3682  0.1821  0.3477  0.3040  0.2917
            usNOAH         0.4015  0.2746  0.3682  1.0000  0.3628  0.2178  0.4645  0.4488
            usSFUN         0.3680  0.1978  0.1821  0.3628  1.0000  0.2513  0.2843  0.4883
            usTSLA         0.3015  0.4080  0.3477  0.2178  0.2513  1.0000  0.2327  0.3340
            usVIPS         0.3706  0.2856  0.3040  0.4645  0.2843  0.2327  1.0000  0.4189
            usWUBA         0.4320  0.2438  0.2917  0.4488  0.4883  0.3340  0.4189  1.0000

            'sperm':
                            usBIDU    usFB  usGOOG  usNOAH  usSFUN  usTSLA  usVIPS  usWUBA
            usBIDU         1.0000  0.3888  0.4549  0.4184  0.3747  0.3623  0.4333  0.4396
            usFB           0.3888  1.0000  0.7013  0.2927  0.2379  0.4200  0.3123  0.2216
            usGOOG         0.4549  0.7013  1.0000  0.3797  0.2413  0.3871  0.3922  0.3035
            usNOAH         0.4184  0.2927  0.3797  1.0000  0.3581  0.2066  0.4643  0.4382
            usSFUN         0.3747  0.2379  0.2413  0.3581  1.0000  0.2645  0.3890  0.4693
            usTSLA         0.3623  0.4200  0.3871  0.2066  0.2645  1.0000  0.2540  0.2801
            usVIPS         0.4333  0.3123  0.3922  0.4643  0.3890  0.2540  1.0000  0.4080
            usWUBA         0.4396  0.2216  0.3035  0.4382  0.4693  0.2801  0.4080  1.0000 }
    """

    if isinstance(corr_jobs, ECoreCorrType):
        # 如果直接传递进来一个ECoreCorrType，暂时兼容，做成序列
        corr_jobs = [corr_jobs]

    if any([not isinstance(corr_job, ECoreCorrType) for corr_job in corr_jobs]):
        # 序列中的所有元素必须是ECoreCorrType
        raise TypeError('corr_job must ECoreCorrType')

    # 匹配市场对应的benchmark
    if ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_US:
        # 美股
        benchmark = IndexSymbol.IXIC
    elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_HK:
        # 港股
        benchmark = IndexSymbol.HSI
    elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_CN:
        # a股
        benchmark = IndexSymbol.SH
    elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN:
        # 国内期货
        benchmark = IndexSymbol.BM_FUTURES_CN
    elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_TC:
        # 币类市场
        benchmark = IndexSymbol.TC_INX
    elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_OPTIONS_US:
        # 美股期权暂时也以IXIC做为标尺
        benchmark = IndexSymbol.IXIC
    elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL:
        # 国际期货暂时也以BM_FUTURES_GB做为标尺
        benchmark = IndexSymbol.BM_FUTURES_GB
    else:
        # 没匹配上也不抛错误，随便给一个，因为这里要的benchmark主要目的只是做为时间标尺
        benchmark = IndexSymbol.IXIC
    # 根据参数使用_all_market_cg获取全市场symbol涨跌幅度pd.DataFrame对象change_df
    change_df = _all_market_cg(benchmark, cmp_cnt=cmp_cnt, n_folds=n_folds, start=start, end=end)

    # 使用corr_jobs个相关系数计算方法分别计算change_df的相关系数，所有结果组成一个字典返回
    return {corr_job.value: ABuCorrcoef.corr_matrix(change_df, corr_job) for corr_job in corr_jobs}



# -*- encoding:utf-8 -*-
"""买入因子参数以及选择图形可视化"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import abstractmethod

import ipywidgets as widgets

from ..WidgetBu.ABuWGBase import WidgetFactorBase, WidgetFactorManagerBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


# noinspection PyUnresolvedReferences
class BFSubscriberMixin(object):
    """
        混入类：订阅买入策略更新通知，以及构建添加基础附属于买入策略的策略
        如：依附于买入策略的卖出策略，依赖于买入策略的选股策略
    """

    def subscriber_ui(self, labels):
        """
        构建订阅的已添加的买入策略ui初始化
        :param labels: list序列内部对象str用来描述解释
        """
        # 添加针对指定买入策略的卖出策略
        self.accordion = widgets.Accordion()
        buy_factors_child = []
        for label in labels:
            buy_factors_child.append(widgets.Label(label,
                                                   layout=widgets.Layout(width='300px', align_items='stretch')))
        self.buy_factors = widgets.SelectMultiple(
            options=[],
            description=u'已添加的买入策略:',
            disabled=False,
            layout=widgets.Layout(width='100%', align_items='stretch')
        )
        buy_factors_child.append(self.buy_factors)
        buy_factors_box = widgets.VBox(buy_factors_child)
        self.accordion.children = [buy_factors_box]

    def notify_subscriber(self):
        """已添加的买入策略框接收买入tab已添加的买入策略信息改变ui同步"""
        self.buy_factors.options = self.wg_manager.buy_factor_manger.selected_factors.options

    def add_to_buy_factor(self, select, factor_unique_callable, factor_key):
        """对应按钮添加策略到指定买入策略中的基础方法，具体策略中外层需要再套一层"""
        if not callable(factor_unique_callable):
            raise TypeError('factor_unique_callable must callable!')

        if self.wg_manager.buy_factor_manger is not None and len(list(select['new'])) > 0:
            #  由于是多选框，简单处理，只能最后一个
            buy_factor_desc_key = list(select['new'])[-1]

            # 构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key
            factor_dict, factor_desc_key = factor_unique_callable()

            desc_key_list = buy_factor_desc_key.split('+')
            if factor_desc_key in desc_key_list:
                # 已添加过相同描述的策略在买入因子中，返回不加了
                return

            # 从买入manager字典中出栈点击的这个描述买入字典对象
            buy_factor_dict = self.wg_manager.buy_factor_manger.factor_dict.pop(buy_factor_desc_key)
            # 买入字典对象中如果有独立因子策略序列，弹出
            factors = buy_factor_dict.pop(factor_key, [])

            # 买入独有因子序列中加入factor_dict
            factors.append(factor_dict)
            # 将因子序列放回到买入字典对象中
            buy_factor_dict[factor_key] = factors
            # 买入策略因子描述＋独有因子描述＝唯一策略描述
            combine_factor_desc_key = u'{}+{}'.format(buy_factor_desc_key, factor_desc_key)
            # 买入策略因子中放回组合好的买入策略
            self.wg_manager.buy_factor_manger.factor_dict[combine_factor_desc_key] = buy_factor_dict
            # 通知ui更新
            self.wg_manager.buy_factor_manger.refresh_factor()


class BuyFactorWGManager(WidgetFactorManagerBase):
    """买入策略组织类"""

    def _init_widget(self):
        """构建内置的买入策略可视化组件，构造出self.factor_box"""

        from ..WidgetBu.ABuWGBuyFactor import BuyDMWidget, BuyXDWidget, BuyWDWidget, BuySDWidget, BuyWMWidget
        self.bf_array = []
        self.bf_array.append(BuyDMWidget(self))
        self.bf_array.append(BuyXDWidget(self))
        self.bf_array.append(BuyWDWidget(self))
        self.bf_array.append(BuySDWidget(self))
        self.bf_array.append(BuyWMWidget(self))

        # bf() call用widget组list
        children = [bf() for bf in self.bf_array]

        if self.scroll_factor_box:
            self.factor_box = widgets.Box(children=children,
                                          layout=self.factor_layout)
        else:
            # 一行显示两个，2个为一组，组装sub_children_group序列,
            sub_children_group = self._sub_children(children, len(children) / 2)
            sub_children_box = [widgets.HBox(sub_children) for sub_children in sub_children_group]
            self.factor_box = widgets.VBox(sub_children_box)


class WidgetFactorBuyBase(WidgetFactorBase):
    """买入策略可视化基础类"""

    def __init__(self, wg_manager):
        super(WidgetFactorBuyBase, self).__init__(wg_manager)
        self.add = widgets.Button(description=u'添加为全局买入策略', layout=widgets.Layout(width='98%'),
                                  button_style='info')
        self.add.on_click(self.add_buy_factor)
        self._init_widget()

    @abstractmethod
    def _init_widget(self):
        """子类因子界面设置初始化"""
        pass

    @abstractmethod
    def make_buy_factor_unique(self):
        """
            子类因子构建唯一描述以及因子字典
            返回值两个：
            1. 因子构建字典对象
            2. 因子唯一描述
        """
        pass

    # noinspection PyUnusedLocal
    def add_buy_factor(self, bt):
        """对应按钮添加策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict, factor_desc_key = self.make_buy_factor_unique()
        self.wg_manager.add_factor(factor_dict, factor_desc_key)



# -*- encoding:utf-8 -*-
"""
    计算线atr模块
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from ..TLineBu.ABuTLine import AbuTLine
from ..CoreBu.ABuPdHelper import pd_rolling_std, pd_ewm_std

__author__ = '阿布'
__weixin__ = 'abu_quant'


def show_atr_std(kl_pd):
    """
    可视化atr移动平均std和加权移动平均std，
    注意会修改kl_pd，只做测试使用，内部未做copy处理，
    如不能改动，外部自copy操作，再传递进来
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    """

    pre_atr21 = kl_pd['atr21'].shift(1)
    # noinspection PyTypeChecker
    kl_pd['atr_change'] = np.where(pre_atr21 == 0, 0, np.log(kl_pd['atr21'] / pre_atr21))
    kl_pd['mov_std'] = pd_rolling_std(kl_pd['atr_change'], window=20, center=False) * np.sqrt(20)
    kl_pd['std_ewm'] = pd_ewm_std(kl_pd['atr_change'], span=20, min_periods=20, adjust=True) * np.sqrt(20)
    kl_pd[['close', 'atr21', 'mov_std', 'std_ewm', 'atr_change']].plot(subplots=True, figsize=(16, 12), grid=True)
    plt.show()


def calc_atr_std(kl_pd, xd=21, ewm=True, show=True):
    """
    计算atr移动平均std或者加权移动平均std技术线，使用
    AbuTLine封装技术线实体，不会修改kl_pd，返回AbuTLine对象
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param xd: 计算移动平均std或加权移动平均std使用的窗口参数，默认21
    :param ewm: 是否使用加权移动平均std计算
    :param show: 是否可视化，可视化使用AbuTLine.show接口
    :return: 返回AbuTLine对象
    """
    pre_atr21 = kl_pd['atr21'].shift(1)
    # noinspection PyTypeChecker
    atr_change = np.where(pre_atr21 == 0, 0, np.log(kl_pd['atr21'] / pre_atr21))

    if ewm:
        atr_roll_std = pd_ewm_std(atr_change, span=xd, min_periods=1, adjust=True) * np.sqrt(xd)
    else:
        atr_roll_std = pd_rolling_std(atr_change, window=xd, min_periods=1, center=False) * np.sqrt(xd)

    # min_periods=1还是会有两个nan，填了
    atr_roll_std = pd.Series(atr_roll_std).fillna(method='bfill')
    # 主要目的就是通过atr_roll_std构造line
    line = AbuTLine(atr_roll_std, 'atr std')
    if show:
        line.show()

    return line



# -*- encoding:utf-8 -*-
"""示例ump边裁价格波动特征模块"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from ..TradeBu.ABuMLFeature import AbuFeatureWave
from ..MLBu.ABuMLPd import AbuMLPd
from .ABuUmpBase import ump_edge_make_xy, BuyUmpMixin
from .ABuUmpEdgeBase import AbuUmpEdgeBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuUmpEdgeWave(AbuUmpEdgeBase, BuyUmpMixin):
    """边裁价格波动特征类，AbuUmpEdgeBase子类，混入BuyUmpMixin，做为买入ump类"""

    class UmpWaveFiter(AbuMLPd):
        """
            内部类，只需要实现make_xy，且使用ump_edge_make_xy装饰

            在边裁__init__中通过：
                    self.fiter_cls = self.get_fiter_class()
                    self.fiter = self.fiter_cls(orders_pd=orders_pd, **kwarg)
            构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
        """

        @ump_edge_make_xy
        def make_xy(self, **kwarg):
            """
             make_xy通过装饰器ump_edge_make_xy进行包装，调用make_xy前将有交易结果的单子进行筛选：
             order_has_ret = orders_pd[orders_pd['result'] != 0]，且赋予self.order_has_ret
             make_xy只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df

            :param kwarg: ump_edge_make_xy装饰器中使用kwarg
                          kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
            :return: self.order_has_ret中使用filter选取特征列形成wave_df
                     ump_edge_make_xy装饰器在make_xy返回wave_df后做转换matrix，形成x，y等工作
            """

            filter_list = ['profit', 'profit_cg']
            cols = AbuFeatureWave().get_feature_ump_keys(ump_cls=AbuUmpEdgeWave)
            # filter_list=['profit', 'profit_cg', 'buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3']
            filter_list.extend(cols)
            # noinspection PyUnresolvedReferences
            wave_df = self.order_has_ret.filter(filter_list)
            """
                eg: wave_df
                               profit  profit_cg  buy_wave_score1  buy_wave_score2  \
                2014-09-24  -22618.04    -0.0566            0.287            0.234
                2014-10-24  -29690.28    -0.0742            0.596            0.488
                2014-10-29   18959.19     0.0542            0.444            0.338
                2014-10-29  148209.36     0.5022           -0.173           -0.202
                2014-10-29   24867.60     0.0952            0.031           -0.128
                2014-10-29   18959.19     0.0542            0.444            0.338
                2014-11-03    1250.80     0.0045            0.018           -0.128
                2014-11-11   59888.21     0.1857           -0.144           -0.060
                2014-11-12   -3578.78    -0.0140           -0.453           -0.505
                2014-11-26  -29085.19    -0.0946           -0.005           -0.007
                ...               ...        ...              ...              ...
                2016-03-14   16220.57     0.0559            0.928            0.941
                2016-03-14  -25328.12    -0.1218            1.209            0.891
                2016-03-30  -29858.44    -0.0863            0.470            0.630
                2016-04-04    5373.76     0.0244            0.363            0.608
                2016-04-13  -28044.40    -0.1159            0.271            0.509
                2016-04-14  -18645.93    -0.0467           -0.030            0.081
                2016-04-15  -32484.79    -0.1149            0.596            0.753
                2016-04-15  -32484.79    -0.1149            0.596            0.753
                2016-04-29     290.96     0.0007            0.743            0.840
                2016-04-29     290.96     0.0007            0.743            0.840

                            buy_wave_score3
                2014-09-24            0.218
                2014-10-24            0.449
                2014-10-29            0.329
                2014-10-29           -0.203
                2014-10-29           -0.173
                2014-10-29            0.329
                2014-11-03           -0.172
                2014-11-11            0.001
                2014-11-12           -0.509
                2014-11-26            0.015
                ...                     ...
                2016-03-14            0.948
                2016-03-14            0.788
                2016-03-30            0.702
                2016-04-04            0.743
                2016-04-13            0.651
                2016-04-14            0.170
                2016-04-15            0.800
                2016-04-15            0.800
                2016-04-29            0.918
                2016-04-29            0.918

            """
            return wave_df

    def get_predict_col(self):
        """
        边裁价格波动特征keys：['buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3']
        :return: ['buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3']
        """
        return AbuFeatureWave().get_feature_ump_keys(ump_cls=AbuUmpEdgeWave)

    def get_fiter_class(self):
        """
        边裁价格波动特征返回的AbuMLPd子类：AbuUmpEdgeWave.UmpWaveFiter
        :return: AbuUmpEdgeWave.UmpWaveFiter
        """
        return AbuUmpEdgeWave.UmpWaveFiter

    @classmethod
    def class_unique_id(cls):
        """
         具体ump类关键字唯一名称，类方法：return 'wave_edge'
         主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
         具体使用见ABuUmpManager中extend_ump_block方法
        """
        return 'wave_edge'



# -*- encoding:utf-8 -*-
"""
    abupy中使用的金融时间序列分析模块, 模块真的方法真的参数都为abupy中格式化好的kl如下

    eg:
                close	high	low	p_change	open	pre_close	volume	date	date_week	key	atr21	atr14
    2016-07-20	228.36	229.800	225.00	1.38	226.47	225.26	2568498	20160720	2	499	9.1923	8.7234
    2016-07-21	220.50	227.847	219.10	-3.44	226.00	228.36	4428651	20160721	3	500	9.1711	8.7251
    2016-07-22	222.27	224.500	218.88	0.80	221.99	220.50	2579692	20160722	4	501	9.1858	8.7790
    2016-07-25	230.01	231.390	221.37	3.48	222.27	222.27	4490683	20160725	0	502	9.2669	8.9298
    2016-07-26	225.93	228.740	225.63	-1.77	227.34	230.01	41833	20160726	1	503	9.1337	8.7541
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import Iterable

import logging

import numpy as np
import pandas as pd

from ..CoreBu import ABuEnv
from ..CoreBu.ABuPdHelper import pd_resample

__author__ = '阿布'
__weixin__ = 'abu_quant'

log_func = logging.info if ABuEnv.g_is_ipython else print


def _df_dispatch(df, dispatch_func):
    """
    根据df的类型分发callable的执行方法，

    :param df: abupy中格式化好的kl，或者字典，或者可迭代序列
    :param dispatch_func: 分发的可执行的方法
    """
    if isinstance(df, pd.DataFrame):
        # 参数只是pd.DataFrame
        return dispatch_func(df)
    elif isinstance(df, dict) and all([isinstance(_df, pd.DataFrame) for _df in df.values()]):
        # 参数只是字典形式
        return [dispatch_func(df[df_key], df_key) for df_key in df]
    elif isinstance(df, Iterable) and all([isinstance(_df, pd.DataFrame) for _df in df]):
        # 参数只是可迭代序列
        return [dispatch_func(_df) for _df in df]
    else:
        log_func('df type is error! {}'.format(type(df)))


def _df_dispatch_concat(df, dispatch_func):
    """
    根据df的类型分发callable的执行方法，如果是字典或者可迭代类型的返回值使用
    pd.concat连接起来

    :param df: abupy中格式化好的kl，或者字典，或者可迭代序列
    :param dispatch_func: 分发的可执行的方法
    """

    if isinstance(df, pd.DataFrame):
        # 参数只是pd.DataFrame
        return dispatch_func(df)
    elif isinstance(df, dict) and all([isinstance(_df, pd.DataFrame) for _df in df.values()]):
        # 参数只是字典形式
        return pd.concat([dispatch_func(df[df_key], df_key) for df_key in df], axis=1)
    elif isinstance(df, Iterable) and all([isinstance(_df, pd.DataFrame) for _df in df]):
        # 参数只是可迭代序列
        return pd.concat([dispatch_func(_df) for _df in df], axis=1)
    else:
        log_func('df type is error! {}'.format(type(df)))


def resample_close_mean(df, bins=None):
    """
    对金融时间序列进行变换周期重新采样，对重新采样的结果进行pct_change处理
    ，对pct_change序列取abs绝对值，对pct_change绝对值序列取平均，即算出
    重新采样的周期内的平均变化幅度

    eg:
    tsla = ABuSymbolPd.make_kl_df('usTSLA')
    ABuKLUtil.resample_close_mean(tsla)

    out:
                resample
        5D	    0.0340
        10D	    0.0468
        21D	    0.0683
        42D	    0.0805
        60D	    0.1002
        90D	    0.0931
        120D    0.0939

    :param df: abupy中格式化好的kl，或者字典，或者可迭代序列
    :param bins: 默认eg:  ['5D', '10D', '21D', '42D', '60D', '90D', '120D']
    :return: pd.DataFrame
    """

    def _resample_close_mean(p_df, df_name=''):
        resample_dict = {}
        for _bin in bins:
            change = abs(pd_resample(p_df.close, _bin, how='mean').pct_change()).mean()
            """
                eg: pd_resample(p_df.close, bin, how='mean')

                    2014-07-23    249.0728
                    2014-09-03    258.3640
                    2014-10-15    240.8663
                    2014-11-26    220.1552
                    2015-01-07    206.0070
                    2015-02-18    198.0932
                    2015-04-01    217.9791
                    2015-05-13    251.3640
                    2015-06-24    266.4511
                    2015-08-05    244.3334
                    2015-09-16    236.2250
                    2015-10-28    222.0441
                    2015-12-09    222.0574
                    2016-01-20    177.2303
                    2016-03-02    226.8766
                    2016-04-13    230.6000
                    2016-05-25    216.7596
                    2016-07-06    222.6420

                    abs(pd_resample(p_df.close, bin, how='mean').pct_change())

                    2014-09-03    0.037
                    2014-10-15    0.068
                    2014-11-26    0.086
                    2015-01-07    0.064
                    2015-02-18    0.038
                    2015-04-01    0.100
                    2015-05-13    0.153
                    2015-06-24    0.060
                    2015-08-05    0.083
                    2015-09-16    0.033
                    2015-10-28    0.060
                    2015-12-09    0.000
                    2016-01-20    0.202
                    2016-03-02    0.280
                    2016-04-13    0.016
                    2016-05-25    0.060
                    2016-07-06    0.027

                    abs(pd_resample(p_df.close, bin, how='mean').pct_change()).mean():

                    0.080
            """
            resample_dict[_bin] = change
        resample_df = pd.DataFrame.from_dict(resample_dict, orient='index')
        resample_df.columns = ['{}resample'.format(df_name)]
        return resample_df

    if bins is None:
        bins = ['5D', '10D', '21D', '42D', '60D', '90D', '120D']
    return _df_dispatch_concat(df, _resample_close_mean)


def bcut_change_vc(df, bins=None):
    """
    eg:
        tsla = ABuSymbolPd.make_kl_df('usTSLA')
        ABuKLUtil.bcut_change_vc(tsla)

        out:
                p_change	rate
        (0, 3]	209	0.4147
        (-3, 0]	193	0.3829
        (3, 7]	47	0.0933
        (-7, -3]	44	0.0873
        (-10, -7]	6	0.0119
        (7, 10]	3	0.0060
        (10, inf]	1	0.0020
        (-inf, -10]	1	0.0020

    :param df: abupy中格式化好的kl，或者字典，或者可迭代序列
    :param bins: 默认eg：[-np.inf, -10, -7, -3, 0, 3, 7, 10, np.inf]
    :return: pd.DataFrame
    """

    def _bcut_change_vc(p_df, df_name=''):
        dww = pd.DataFrame(pd.cut(p_df.p_change, bins=bins).value_counts())
        # 计算各个bin所占的百分比
        dww['{}rate'.format(df_name)] = dww.p_change.values / dww.p_change.values.sum()
        if len(df_name) > 0:
            dww.rename(columns={'p_change': '{}'.format(df_name)}, inplace=True)
        return dww

    if bins is None:
        bins = [-np.inf, -10, -7, -3, 0, 3, 7, 10, np.inf]
    return _df_dispatch_concat(df, _bcut_change_vc)


def qcut_change_vc(df, q=10):
    """
    eg:
        tsla = ABuSymbolPd.make_kl_df('usTSLA')
        ABuKLUtil.qcut_change_vc(tsla)

        out:
            change
        0	[-10.45, -3.002]
        1	(-3.002, -1.666]
        2	(-1.666, -0.93]
        3	(-0.93, -0.396]
        4	(-0.396, 0.065]
        5	(0.065, 0.48]
        6	(0.48, 1.102]
        7	(1.102, 1.922]
        8	(1.922, 3.007]
        9	(3.007, 11.17]

    :param df: abupy中格式化好的kl，或者字典，或者可迭代序列
    :param q: 透传qcut使用的q参数，默认10，10等分
    :return: pd.DataFrame
    """

    def _qcut_change_vc(p_df, df_name=''):
        dww = pd.qcut(p_df.p_change, q).value_counts().index.values
        # 构造Categories使用DataFrame套Series
        dww = pd.Series(dww)
        # 涨跌从负向正开始排序
        dww.sort_values(inplace=True)
        dww = pd.DataFrame(dww)
        # 排序后index重新从0开始排列
        dww.index = np.arange(0, q)
        dww.columns = ['{}change'.format(df_name)]
        return dww

    return _df_dispatch_concat(df, _qcut_change_vc)


def date_week_mean(df):
    """
        eg:

        tsla = ABuSymbolPd.make_kl_df('usTSLA')
        ABuKLUtil.date_week_mean(tsla)

        out:
        周一    0.0626
        周二    0.0475
        周三    0.0881
        周四    0.2691
        周五   -0.2838
    :param df: abupy中格式化好的kl，或者字典，或者可迭代序列
    :return: pd.Series或者pd.DataFrame
    """

    def _date_week_win(p_df, df_name=''):
        dww = p_df.groupby('date_week')['p_change'].mean()
        # 将周几这个信息变成中文
        dww.rename(index={6: '周日', 0: '周一', 1: '周二', 2: '周三', 3: '周四', 4: '周五', 5: '周六'},
                   inplace=True)
        # p_change变成对应的pchange
        dww = pd.DataFrame(dww)
        dww.rename(columns={'p_change': '{}_p_change'.format(df_name)}, inplace=True)
        return dww

    return _df_dispatch_concat(df, _date_week_win)


def date_week_win(df):
    """
    eg:
        tsla = ABuSymbolPd.make_kl_df('usTSLA')
        ABuKLUtil.date_week_win(tsla)

        out：
                0	1	win
            date_week
            0	44	51	0.5368
            1	55	48	0.4660
            2	48	57	0.5429
            3	44	57	0.5644
            4	53	47	0.470

    :param df: bupy中格式化好的kl，或者字典，或者可迭代序列
    :return: pd.Series或者pd.DataFrame
    """

    def _date_week_win(p_df, df_name=''):
        _df = p_df.copy()
        win_key = '{}win'.format(df_name)
        _df[win_key] = _df['p_change'] > 0
        _df[win_key] = _df[win_key].astype(int)

        dww = pd.concat([pd.crosstab(_df.date_week, _df[win_key]), _df.pivot_table([win_key], index='date_week')],
                        axis=1)
        # 将周几这个信息变成中文
        # noinspection PyUnresolvedReferences
        dww.rename(index={6: '周日', 0: '周一', 1: '周二', 2: '周三', 3: '周四', 4: '周五', 5: '周六'}, inplace=True)
        return dww

    return _df_dispatch_concat(df, _date_week_win)


def wave_change_rate(df):
    """
    eg:
        tsla = ABuSymbolPd.make_kl_df('usTSLA')
        ABuKLUtil.wave_change_rate(tsla)

        out:
        日振幅涨跌幅比：1.794156

    :param df: abupy中格式化好的kl，或者字典，或者可迭代序列
    """

    def _wave_change_rate(p_df, df_name=''):
        wave = ((p_df.high - p_df.low) / p_df.pre_close) * 100
        # noinspection PyUnresolvedReferences
        wave_rate = wave.mean() / np.abs(p_df['p_change']).mean()

        print('{}日振幅涨跌幅比：{:2f}, {}日统计套利条件'.format(
            df_name, wave_rate, '具备' if wave_rate > 1.80 else '不具备'))
    _df_dispatch(df, _wave_change_rate)


def p_change_stats(df):
    """
    eg :
        tsla = ABuSymbolPd.make_kl_df('usTSLA')
        ABuKLUtil.p_change_stats(tsla)

        out:

        日涨幅平均值1.861, 共260个交易日上涨走势
        日跌幅平均值-1.906, 共244个交易日下跌走势
        日平均涨跌比0.977, 上涨下跌数量比:1.066

    :param df: abupy中格式化好的kl，或者字典，或者可迭代序列
    """

    def _p_change_stats(p_df, df_name=''):
        p_change_up = p_df[p_df['p_change'] > 0].p_change
        p_change_down = p_df[p_df['p_change'] < 0].p_change
        print('{}日涨幅平均值{:.3f}, 共{}个交易日上涨走势'.format(df_name, p_change_up.mean(), p_change_up.count()))
        print('{}日跌幅平均值{:.3f}, 共{}个交易日下跌走势'.format(df_name, p_change_down.mean(), p_change_down.count()))
        print('{}日平均涨跌比{:.3f}, 上涨下跌数量比:{:.3f}\n'.format(
            df_name, abs(p_change_up.mean() / p_change_down.mean()), p_change_up.count() / p_change_down.count()))

    _df_dispatch(df, _p_change_stats)


def date_week_wave(df):
    """
    根据周几分析金融时间序列中的日波动:

    eg:
        tsla = ABuSymbolPd.make_kl_df('usTSLA')
        ABuKLUtil.date_week_wave(tsla)

        out:
            usTSLAwave
            date_week
            周一  3.8144
            周二  3.3326
            周三  3.3932
            周四  3.3801
            周五  2.9923

    :param df: abupy中格式化好的kl，或者字典，或者可迭代序列
    :return: pd.Series或者pd.DataFrame
    """

    def _date_week_wave(p_df, df_name=''):
        # 要改df所以copy
        df_copy = p_df.copy()
        wave_key = '{}wave'.format(df_name)
        # 计算波动: * 100目的是和金融序列中的p_change单位一致
        df_copy[wave_key] = ((df_copy.high - df_copy.low) / df_copy.pre_close) * 100
        dww = df_copy.groupby('date_week')[wave_key].mean()
        # 将周几这个信息变成中文
        dww.rename(index={6: '周日', 0: '周一', 1: '周二', 2: '周三', 3: '周四', 4: '周五', 5: '周六'}, inplace=True)
        return dww

    return _df_dispatch_concat(df, _date_week_wave)
